/*
 Copyright 2011-2018 Fulco Scherjon

This file is part of the HomininSpace modelling and simulation system.

HomininSpace is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, any version 3.
HomininSpace is distributed WITHOUT ANY WARRANTY under the GNU General Public License.
HomininSpace comes with a copy of this GNU General Public License.  If not, see <http://www.gnu.org/licenses/>.

*/

package HomininSpace;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.logging.ErrorManager;
import java.util.logging.FileHandler;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;
import java.util.logging.SimpleFormatter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import repast.simphony.context.Context;
import repast.simphony.context.DefaultContext;
import repast.simphony.context.space.continuous.ContinuousSpaceFactoryFinder;
import repast.simphony.context.space.grid.GridFactoryFinder;
import repast.simphony.dataLoader.ContextBuilder;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunState;
import repast.simphony.parameter.Parameters;
import repast.simphony.query.space.grid.GridCell;
import repast.simphony.query.space.grid.GridCellNgh;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.continuous.ContinuousSpace;
import repast.simphony.space.continuous.RandomCartesianAdder;
import repast.simphony.space.grid.Grid;
import repast.simphony.space.grid.GridBuilderParameters;
import repast.simphony.space.grid.GridPoint;
import repast.simphony.space.grid.SimpleGridAdder;
import repast.simphony.space.grid.WrapAroundBorders;
import repast.simphony.util.SimUtilities;
import repast.simphony.valueLayer.GridValueLayer;
import repast.simphony.engine.schedule.ISchedule;
import repast.simphony.engine.schedule.ScheduleParameters;
import HomininSpace.common.*;
import HomininSpace.agents.*;

/*
 * MyTextFormatter, log record formatter 
 */
	//This custom formatter formats parts of a log record to a single line
	class MyTextFormatter extends Formatter 
	{
	  // This method is called for every log records
	  public String format(LogRecord rec) 
	  {
	      StringBuffer buf = new StringBuffer(1000);
	      buf.append(formatMessage(rec));
	      buf.append('\n');
	      return buf.toString();
	  }
	
	  // This method is called just after the handler using this formatter is created
	  public String getHead(Handler h) 
	  {
	      return "HomininSpace Simulation "+(new Date())+"\n";
	  }
	
	  // This method is called just after the handler using this formatter is closed
	  public String getTail(Handler h) 
	  {
	      return "End\n";
	  }
	}  // MyTextFormatter

/**
 * Custom {@link ContextBuilder} implementation for the HomininSpace modelling environment.
 * 
 */
public class HomininSpaceContextBuilder extends DefaultContext<Object> implements
		ContextBuilder<Object> {
	
	private static int 							agentID = 1; // Used to generate unique agent ids
	private static double 						startTime;
	private static boolean						simulationHasEnded = false;
	private static int 							tickCount = 0;
	private static Logger 						logger = null;
	private static FileHandler 					logFileHandler = null;
	private Parameters 							parameters = null;
    private String		 						parameterLine = null;		// the original parameters for this scenario
    private String		 						parameterHeader = null;		// the line with the names of the parameters
	private static int							randomSeed;
	private static String						timeStampName;
	
	private static int							simulationId = 0;
	
	private static ArrayList<HomininGroup> 		deadHomininList = null;
	private static ArrayList<HomininFactory> 	factoryList = null;
	private static ArrayList<HabitatCell> 		habitatList = null;
	private static ArrayList<Checkpoint> 		checkpointList = null;
	private static ArrayList<Checkpoint> 		climatecheckpointList = null;
	
	private static Grid<Object> 				theGrid = null; 
	private static Context<Object> 				theContext = null;
	private GridValueLayer 						visibleValueLayer = null;
	private static int							gridSizeX = 0;
	private static int 							gridSizeY = 0;
	
	private static double 						globalSealevel = 0.0;
	private static double 						globalTemperatureOffset = 0.0;
	
	private static double						hungryGroupPercentage = 0;
	private static double						hungryHomininPercentage = 0.0;

	// globals constant variables, read from parameters
	private static double						minimumRequiredTemperature = 0.0;
	private static int							birthsPercentage = 33;
	private static int 							sizePreFertileGroup		= 0;
	private static int 							yearsBeforeGroupMaturity		= 0;
	private static int 							sizeFertileGroup		= 0;
	private static boolean						openBorders = false;   // if true groups are removed from the simulation when passing a border
	private static boolean						staticDispersal = Constants.STATIC_DISPERSAL;  // if true groups can settle
	private static boolean						pauzeBeforeEnd = false;
	private static int 							deathsPreFertilePercentage		= 0;
	private static int 							deathsFertilePercentage		= 0;
	private static int 							deathsPostFertilePercentage		= 0;
	private static int 							consumptionPreFertile		= 0;
	private static int 							consumptionFertile		= 0;
	private static int 							consumptionPostFertile		= 0;
	private static int 							groupSizeLarge		= 0;
	private static int 							groupSizeSmall		= 0;
	private static int 							groupSizeFertileSmall		= 0;
	
	private static int 							caloriesPerKgMeat		= 0;
	private static int 							maxForagingRange		= 0;
	private static int 							precipitationColdArid		= 0;
	private static int 							precipitationWarmArid		= 0;
	private static int 							biomassRecoverPercentage		= 0;
	
	private static boolean						useFactories = true;
	private static int 							intialIberiaGroups = 0;
	private static int 							intialItalyGroups = 0;
	
	private static String						modelDataDirectory = "";
	
	public static final ContinuousSpace<Object> theSpace = null;
	
	/*
	 * Write parameter values to logfile
	 */
	private void logParameters()
	{
        // log parameters
        logger.info("startingTimeStep: " + HSUtils.startingTimeStep );
        logger.info("timeStepIncrement: " + HSUtils.timeStepIncrement  );
        logger.info("DEBUG_LEVEL: " + Constants.DEBUG_LEVEL );
        logger.info("tickCount: " + tickCount );
        logger.info("minimumRequiredTemperature: " + minimumRequiredTemperature );
        logger.info("birthsPercentage: " + birthsPercentage );
        logger.info("sizePreFertileGroup: " + sizePreFertileGroup );
        logger.info("sizeFertileGroup: " + sizeFertileGroup );
        logger.info("deathsPreFertilePercentage: " + deathsPreFertilePercentage );
        logger.info("deathsFertilePercentage: " + deathsFertilePercentage );
        logger.info("deathsPostFertilePercentage: " + deathsPostFertilePercentage );
        logger.info("consumptionPreFertile: " + consumptionPreFertile );
        logger.info("consumptionFertile: " + consumptionFertile );
        logger.info("consumptionPostFertile: " + consumptionPostFertile );
        logger.info("yearsBeforeGroupMaturity: " + yearsBeforeGroupMaturity );
        logger.info("groupSizeLarge: " + groupSizeLarge );
        logger.info("groupSizeSmall: " + groupSizeSmall );
        logger.info("groupSizeFertileSmall: " + groupSizeFertileSmall );
        logger.info("caloriesPerKgMeat: " + caloriesPerKgMeat );
        logger.info("maxForagingRange: " + maxForagingRange );
        logger.info("precipitationColdArid: " + precipitationColdArid );
        logger.info("precipitationWarmArid: " + precipitationWarmArid );
        logger.info("biomassRecoverPercentage: " + biomassRecoverPercentage );
        logger.info("openBorders: " + openBorders );
        logger.info("pauzeBeforeEnd: " + pauzeBeforeEnd );
        logger.info("staticDispersal: " + staticDispersal );
        logger.info("useFactories: " + useFactories );
        logger.info("intialIberiaGroups: " + intialIberiaGroups );
        logger.info("intialItalyGroups: " + intialItalyGroups );
        
		logger.info("Ordering: Local, Iberia, Italy.");
		if (Constants.ENERGY_CONTINUOUS)
		{
			logger.info("Continuous energy model (McNaughton)");
		}
		else
		{
			logger.info("Habitat model (Kelly/Binford)");
		}
        
		logger.info("Random seed: " + randomSeed);
		if (Constants.ZERO_MODEL_RANDOM_WALK)   // random movement
		{
			logger.info("RANDOM walk activated");
		}

	}  // logParameters
	
	
	/*
	 * Constructor, build simulation environment
	 */
	@Override
	public Context<Object> build(final Context<Object> aContext) 
	{
		int		settingsIndex;
		HomininGroup hominin;
		int homininCount = 				0;
		
		simulationHasEnded = 			false;
		theContext = 					aContext;
		
		deadHomininList = 				new ArrayList<HomininGroup>();

		// Set a specified context ID
		theContext.setId(Constants.CONTEXT_ID);

        HSUtils.initialize();
        
        // set default values for global parameters
        minimumRequiredTemperature = 	-18;
        birthsPercentage = 				33;
		sizePreFertileGroup = 			15;		
		sizeFertileGroup = 				25;
		deathsPreFertilePercentage = 	4;
		deathsFertilePercentage = 		2;
		deathsPostFertilePercentage = 	8;
		consumptionPreFertile = 		3000;
		consumptionFertile = 			4000;
		consumptionPostFertile = 		3500;
		yearsBeforeGroupMaturity = 		5;
		groupSizeLarge = 				50;
		groupSizeSmall = 				10;
		groupSizeFertileSmall = 		3;
		caloriesPerKgMeat = 			3000;
		maxForagingRange =				5;
		precipitationColdArid = 		609;
		precipitationWarmArid = 		974;
		biomassRecoverPercentage = 		20;
		
		// simulation system parameters
        openBorders = 					false;
        pauzeBeforeEnd = 				false;
        HSUtils.startingTimeStep = 		(int) Math.round(Constants.START_TIME_STEP*1000);
        HSUtils.timeStepIncrement = 	1;
        Constants.DEBUG_LEVEL = 		10;
		tickCount = 					1000 * ((int) Math.round(Constants.START_TIME_STEP-Constants.END_TIME_STEP));  // was: 81.000
		HSUtils.numberOftimeSteps = 	tickCount;
        
        useFactories = 					true;
        intialIberiaGroups = 			0;
        intialItalyGroups = 			0;

        modelDataDirectory = 			Constants.MODEL_DATA_DIRECTORY;

		// Get any parameter values from options (user interface)
        parameters = RunEnvironment.getInstance().getParameters();
        Constants.DEBUG_LEVEL = ((Integer) parameters.getValue(Constants.PARAMETER_ID_DEFAULT_LEVEL)).intValue();
        settingsIndex = ((Integer) parameters.getValue(Constants.PARAMETER_ID_SETTINGS_INDEX)).intValue();

        // read values from user interface parameters (old code, no longer implemented but replaced by input file)
/*        
        HSUtils.startingTimeStep = ((Integer) parameters.getValue(Constants.PARAMETER_ID_STARTING_TIME_STEP)).intValue();
        HSUtils.timeStepIncrement = ((Integer) parameters.getValue(Constants.PARAMETER_ID_TIME_STEP_INCREMENT)).intValue();
        Constants.DEBUG_LEVEL = ((Integer) parameters.getValue(Constants.PARAMETER_ID_DEFAULT_LEVEL)).intValue();
		tickCount = ((Integer) parameters.getValue(Constants.PARAMETER_ID_TICK_COUNT)).intValue();
		HSUtils.numberOftimeSteps = tickCount;
        createRandomSeed = ((Boolean) parameters.getValue(Constants.PARAMETER_ID_CREATE_RANDOM_SEED)).booleanValue();

        minimumRequiredTemperature = ((Double) parameters.getValue(Constants.PARAMETER_ID_MINIMUM_TEMPERATURE)).doubleValue();
        birthsPercentage = ((Integer) parameters.getValue(Constants.PARAMETER_ID_BIRTHS_PERCENTAGE)).intValue();
		sizePreFertileGroup = ((Integer) parameters.getValue("sizePreFertileGroup")).intValue();
		sizeFertileGroup = ((Integer) parameters.getValue("sizeFertileGroup")).intValue();
		deathsPreFertilePercentage = ((Integer) parameters.getValue("deathsPreFertilePercentage")).intValue();
		deathsFertilePercentage = ((Integer) parameters.getValue("deathsFertilePercentage")).intValue();
		deathsPostFertilePercentage = ((Integer) parameters.getValue("deathsPostFertilePercentage")).intValue();
		consumptionPreFertile = ((Integer) parameters.getValue("consumptionPreFertile")).intValue();
		consumptionFertile = ((Integer) parameters.getValue("consumptionFertile")).intValue();
		consumptionPostFertile = ((Integer) parameters.getValue("consumptionPostFertile")).intValue();
		groupSizeLarge = ((Integer) parameters.getValue("groupSizeLarge")).intValue();
		
		caloriesPerKgMeat = ((Integer) parameters.getValue("caloriesPerKgMeat")).intValue();
		precipitationColdArid = ((Integer) parameters.getValue("precipitationColdArid")).intValue();
		precipitationWarmArid = ((Integer) parameters.getValue("precipitationWarmArid")).intValue();
		biomassRecoverPercentage = ((Integer) parameters.getValue("biomassRecoverPercentage")).intValue();
		
        openBorders = ((Boolean) parameters.getValue(Constants.PARAMETER_ID_OPEN_BORDERS)).booleanValue();
        pauzeBeforeEnd = ((Boolean) parameters.getValue(Constants.PARAMETER_ID_PAUZE_BEFORE_END)).booleanValue();
        staticDispersal = ((Boolean) parameters.getValue(Constants.PARAMETER_ID_STATIC_DISPERSAL)).booleanValue();
        
        useFactories = ((Boolean) parameters.getValue("useFactories")).booleanValue();
        factoriesCheckHungryGroups = ((Boolean) parameters.getValue("factoriesCheckHungryGroups")).booleanValue();
        factoriesCheckDensity = ((Boolean) parameters.getValue("factoriesCheckDensity")).booleanValue();
        maximumHungryGroupPercentage = ((Integer) parameters.getValue(Constants.PARAMETER_ID_FACTORY_HUNGER_PERCENTAGE)).intValue();
        intialIberiaGroups = ((Integer) parameters.getValue(Constants.PARAMETER_ID_INITIAL_IBERIA_GROUPS)).intValue();
        intialItalyGroups = ((Integer) parameters.getValue(Constants.PARAMETER_ID_INITIAL_ITALY_GROUPS)).intValue();
*/        
        
		// for logging messages to the console in a specific format
		Handler consoleHandler = new Handler()
		{
		    @Override
		    public void publish(LogRecord record)
		    {
		        if (getFormatter() == null)
		        {
		            setFormatter(new SimpleFormatter());
		        }

		        try {
		            String message = getFormatter().format(record);
		            if (record.getLevel().intValue() >= Level.WARNING.intValue())
		            {
		                System.err.write(message.getBytes());                       
		            }
		            else
		            {
		                System.out.write(message.getBytes());
		            }
		        } catch (Exception exception) {
		            reportError(null, exception, ErrorManager.FORMAT_FAILURE);
		            return;
		        }
		    }

		    @Override
		    public void close() throws SecurityException {}
		    @Override
		    public void flush(){}
		};
		consoleHandler.setFormatter(new MyTextFormatter());
		
		// setup log information, to file and console
		logger = Logger.getLogger("HomininSpaceLogger");
		logFileHandler = null;
		try {
			final String DATE_FORMAT_NOW = "yyyy-MM-dd HHmm";  // remove -ss because we assume that within one minute all checkpoints are initialized, and thus use the same file
		    Calendar cal = Calendar.getInstance();
			SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT_NOW);
			timeStampName = sdf.format(cal.getTime());
			String fileName = "HS Logfile " + timeStampName + ".txt";
			logFileHandler = new FileHandler(fileName);
			logFileHandler.setFormatter(new MyTextFormatter());
		} catch (SecurityException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		logger.setUseParentHandlers(false);
		logger.addHandler(logFileHandler);  				// for logging to file
		logger.addHandler(consoleHandler);   	// for logging to the console

		// Request that every detail gets logged.
        logger.setLevel(Level.ALL);
        
        // INFO message goes on screen and into file
        logger.info("HomininSpace system initialized.");
		logger.info("SettingsIndex: " + settingsIndex );

		/*
		 * Read parameter setting information from file, at position SettingsIndex
		 */
        BufferedReader 				br = null;
        int 						lineCount = 0;
        String						settingsFile = Constants.MODEL_DATA_DIRECTORY + "HomininSpace Scenario Settings.txt";
        
        try {
            br = new BufferedReader(new FileReader(settingsFile));

            // Skip header first line
            parameterHeader = br.readLine();
            int dummyInt;

            while ((parameterLine = br.readLine()) != null) 
            {
                lineCount++;

                // Look for the indicated parameter string (scenario)
                if (lineCount == settingsIndex)
                {
	                // Split the line around separation character (tab, comma)
	                final String[] data = parameterLine.split(",");
	                int idx = 0;
	                simulationId = Integer.parseInt(data[idx++]);
	                
	                	// this is the parameterstring and scenarionumber for this simulation
	                	logger.info("Settings: " + parameterLine );
	                	dummyInt = Integer.parseInt(data[idx++]);
	                   	randomSeed = dummyInt;
	                   	birthsPercentage = Integer.parseInt(data[idx++]);
	            		deathsPreFertilePercentage = Integer.parseInt(data[idx++]);
	            		deathsFertilePercentage = Integer.parseInt(data[idx++]);
	            		deathsPostFertilePercentage = Integer.parseInt(data[idx++]);
	            		consumptionPreFertile = Integer.parseInt(data[idx++]);
	            		consumptionFertile = Integer.parseInt(data[idx++]);
	            		consumptionPostFertile = Integer.parseInt(data[idx++]);
	            		yearsBeforeGroupMaturity = Integer.parseInt(data[idx++]); 
	                	groupSizeSmall = Integer.parseInt(data[idx++]);
	                	groupSizeFertileSmall = Integer.parseInt(data[idx++]);
	                	groupSizeLarge = Integer.parseInt(data[idx++]);
	                	minimumRequiredTemperature  = Integer.parseInt(data[idx++]);
	            		sizePreFertileGroup = Integer.parseInt(data[idx++]);
	            		sizeFertileGroup = Integer.parseInt(data[idx++]);
	            		caloriesPerKgMeat = Integer.parseInt(data[idx++]);
	            		maxForagingRange = Integer.parseInt(data[idx++]);
	            		
	            		// ignore rest for now

	            		break;
                }
            }
            
            Constants.DEBUG_PRINT_LEVEL("Read settings from input file, line#:" + lineCount, 10);
    		logger.info("SimulationId: " + simulationId);
 
        } catch (final FileNotFoundException e) {
                e.printStackTrace();
                return null;
        } catch (final IOException e) {
                e.printStackTrace();
                return null;
        } catch (final NumberFormatException e) {
            e.printStackTrace();
            return null;
        } finally {
                if (br != null) {
                        try {
                                br.close();
                        } catch (final IOException e) {
                                e.printStackTrace();
                        }
                }
        }

        if (simulationId == 0)
        {
        	// failed somehow to read data, return without starting any simulation
        	return null;
        }
        
        logParameters();
        
        // set random seed
		RandomHelper.setSeed(randomSeed);
		RandomHelper.createUniform(0,1);
		logger.info("Random seed: " + randomSeed );
		
        /*
         * read geographic data from specified file
         */
		logger.info("Reading Climate data file: HomininSpace Climate Grid.txt");
		final List<CellData> cellData = HSUtils.readClimateDataFile(modelDataDirectory + "HomininSpace Climate Grid.txt");

        /*
         * read sea levels and global temperatures
         */
		logger.info("Reading sealevel file: bintanja2008.txt");
        HSUtils.readBintanjeDataFile(modelDataDirectory + "bintanja2008.txt");

        // The +1s are added because the grid is indexed from 0, so we have to
        // increase the actual size by 1
        gridSizeX = Collections.max(cellData,
                        CellData.CELL_DATA_X_COMPARATOR).getX() + 1;
        gridSizeY = Collections.max(cellData,
                        CellData.CELL_DATA_Y_COMPARATOR).getY() + 1;

		logger.info("Creating grid: " + gridSizeX + "x" + gridSizeY);
		Constants.DEBUG_PRINT_LEVEL("gridX = " + gridSizeX + " gridY = " + gridSizeY, 10);

		globalSealevel = HSUtils.getSeaLevel((int)Math.round(getTickCount()));
		globalTemperatureOffset = HSUtils.getTemperature((int)Math.round(getTickCount()));
		
		logger.info("Starting year = "+ HSUtils.startingTimeStep + " years BP; sealevel: " + getGlobalSeaLevel() + "; temperature offset: " + getGlobalTemperatureOffset());
       
		// Create a toridal space with random positioning within the specified dimensions
		final ContinuousSpace<Object> space = ContinuousSpaceFactoryFinder
				.createContinuousSpaceFactory(null) // No hints
				.createContinuousSpace(
						Constants.SPACE_ID,
						theContext,
						new RandomCartesianAdder<Object>(),
						new repast.simphony.space.continuous.WrapAroundBorders(),
						gridSizeX, gridSizeY);
		if (space == null)
		{
			// this check on space prevents compiler warning
			logger.info("Unable to create space");
		}
		
		// Create the default grid to display where agents move
		if (getOpenBorders() == true)
		{
			theGrid = GridFactoryFinder
				.createGridFactory(null)
				.createGrid(
						Constants.GRID_ID,
						theContext,
						new GridBuilderParameters<Object>(
//								new repast.simphony.space.grid.BouncyBorders(),  // TODO better?
								new repast.simphony.space.grid.StickyBorders(),
								
// 								WrapAroundBorders which wraps the
//								borders, StrictBorders which throws an exception if an invalid
//								coordinate were specified, BouncyBorders which bounces back the agents
//								from the borders, StickyBorders which don't let the agents cross the
//								borders, etc.)	
								
								// This is a simple implementation of an adder
								// that doesn't perform any action
								new SimpleGridAdder<Object>(),
								// Each cell in the grid is multi-occupancy from now
								true,
								// Size of the grid (defined constants)
								gridSizeX, gridSizeY));
		}
		else
		{
			theGrid = GridFactoryFinder
					.createGridFactory(null)
					.createGrid(
							Constants.GRID_ID,
							theContext,
							new GridBuilderParameters<Object>(
									new repast.simphony.space.grid.BouncyBorders(),  // TODO better?
//									new repast.simphony.space.grid.StickyBorders(),
									
//	 								WrapAroundBorders which wraps the
//									borders, StrictBorders which throws an exception if an invalid
//									coordinate were specified, BouncyBorders which bounces back the agents
//									from the borders, StickyBorders which don't let the agents cross the
//									borders, etc.)	
									
									// This is a simple implementation of an adder
									// that doesn't perform any action
									new SimpleGridAdder<Object>(),
									// Each cell in the grid is multi-occupancy from now
									true,
									// Size of the grid (defined constants)
									gridSizeX, gridSizeY));
		}
		
		/*
		 * Create visual layer
		 */
		visibleValueLayer = new GridValueLayer(
                Constants.VISIBLE_VALUE_LAYER_ID,
                true,
                new WrapAroundBorders(),
                gridSizeX, gridSizeY);
        theContext.addValueLayer(visibleValueLayer);

        /*
         * Create HabitatCell for each CellData object
         */
        for (final CellData data : cellData) 
        {
            final HabitatCell habitatCell = new HabitatCell(data);
            theContext.add(habitatCell);
            
            // first timestep show topography in the visual layer
            visibleValueLayer.set(habitatCell.getHeigth()+globalSealevel, habitatCell.getXPos(), habitatCell.getYPos());
            theGrid.moveTo(habitatCell, data.getX(), data.getY());
        }
        
        // fill climate parameters for those cells that did not get any
        ExtendLandClimate();
        
		/*
		 * Create factories
		 */
       
		logger.info("Creating factories for Iberia (6,9) and Italy (125,20)");
		
        // Iberia
        HomininFactory factory = new HomininFactory(Constants.CULTURAL_TYPE_1);
        factory.setName("Iberia");
        theContext.add(factory);
		factory.initialize( theGrid );
		factory.setPosXY(6, 9);
		factory.setCanSettle(staticDispersal);
		
		// Italy
        factory = new HomininFactory(Constants.CULTURAL_TYPE_2);
        factory.setName("Italy");
        theContext.add(factory);
		factory.initialize( theGrid );
		factory.setPosXY(125, 20);
		factory.setCanSettle(staticDispersal);
		
		/*
		 * Create checkpoints
		 */
		
        /*
         * Read checkpoint data from specified file
         */
		logger.info("Reading checkpoints: HomininSpace Checkpoint List.txt");
        final List<Checkpoint> checkpointData = HSUtils.readCheckpointDataFile(modelDataDirectory + "HomininSpace Checkpoint List.txt");
        int initialHominins = 0;
        int checkpointCount = 0;
        
        for (final Checkpoint cst : checkpointData) 
        {
        	checkpointCount++;
			theContext.add(cst);
			cst.initialize( theGrid );
			cst.setPosXY(cst.getPosX(), cst.getPosY());  // this moves onto the grid
			
            final Iterable<Object> objects = theGrid.getObjectsAt(cst.getPosX(),cst.getPosY());
            for (final Object object : objects) 
            {
            	if (object instanceof HabitatCell) 
            	{
            		HabitatCell h = (HabitatCell) object;
            		h.addCheckpoint(cst);
            		cst.addHabitatCell(h);
            		logger.info("Added checkpoint " + cst);
            		break;
            	}
            }

            // Create initial hominin groups for each starting checkpoint
            if (cst.getType() == Constants.CheckpointType.CHECKPOINT_TYPE_STARTINGPOINT_CST)
			{
            	initialHominins++;
            	hominin = new HomininGroup(theGrid);
				theContext.add(hominin);
				hominin.initialize( homininCount++ );
				hominin.setPosXY(cst.getPosX(), cst.getPosY(), 0, 0);
				hominin.setSize(Constants.INITIAL_SIZE_PREFERTILE, Constants.INITIAL_SIZE_FERTILE, Constants.INITIAL_SIZE_POSTFERTILE);
				hominin.setGroupSettled(true && getStaticDispersal());
				hominin.setHomininType(Constants.CULTURAL_TYPE_0);
			}


        }
		logger.info("Added " + checkpointCount + " checkpoints and " + initialHominins + " initial hominin groups.");
        
        
		// schedule start and end of simulation
		ISchedule schedule = RunEnvironment.getInstance().getCurrentSchedule();
		// schedule last routines to be executed
		ScheduleParameters endParams = ScheduleParameters.createAtEnd(ScheduleParameters.LAST_PRIORITY);
		schedule.schedule(endParams, this, "end");
		// schedule start
		ScheduleParameters startParams = ScheduleParameters.createOneTime(1);
		schedule.schedule(startParams, this, "start");
		// schedule end
        RunEnvironment.getInstance().endAt(tickCount);

        if (pauzeBeforeEnd)
        {
        	// schedule pause, to show topography just before end (density is shown at the final end step)
        	RunEnvironment.getInstance().pauseAt(tickCount-1);
        }
        
        // schedule main loop
		ScheduleParameters loopParams = ScheduleParameters.createRepeating(1, 1, ScheduleParameters.LAST_PRIORITY);
		schedule.schedule(loopParams, this, "activateAgents");

		Constants.DEBUG_PRINT_LEVEL("Sealevel: " + HSUtils.getSeaLevel(0) + " temp: " + HSUtils.getTemperature(0), 10);
		Constants.DEBUG_PRINT_LEVEL("HomininSpace Modelling and Simulation system initialised", 10);

		return theContext;
	}
	
	/*
	 * Start of simulation
	 */
	public void start() 
	{
		startTime = System.currentTimeMillis();
		logger.info("Simulation started");
		
		// Update HabitatCells with initial production
		habitatList = getHabitatCellList();
	    for (final HabitatCell habitat : habitatList) 
	    {
	    	habitat.initialize();
        	habitat.updateHabitatCell();
        	
        	// set a list of Habitatcells as the environment of each h (range = 1) (for avoidance of deathtraps)
    		GridPoint habitatLocation = theGrid.getLocation(habitat);
    		final List<GridCell<HabitatCell>> theNeighborhood = new GridCellNgh<HabitatCell>(theGrid,
    				habitatLocation, HabitatCell.class, 1, 1).getNeighborhood(false);
    		habitat.setEnvironmentList(theNeighborhood);
        	
	    }

	    // Produce initial factory groups
	    // for Iberia
        factoryList = getFactoryList();
		while (intialIberiaGroups > 0)
		{
			intialIberiaGroups--;
	        for (final HomininFactory factory : factoryList) 
	        {
	        	if (factory.getName().equals("Iberia"))
	        	{
	        		// forced production
        		   	factory.produce(true);
	        	}
	        }
		}
		// for Italy
		while (intialItalyGroups > 0)
		{
			intialItalyGroups--;
	        for (final HomininFactory factory : factoryList) 
	        {
	        	if (factory.getName().equals("Italy"))
	        	{
	        		// forced production
        		   	factory.produce(true);
	        	}
	        }
		}
		
		checkpointList = getCheckpointList();
		climatecheckpointList = getClimateCheckpointList();
		
	}	// start

	/*
	 * End of simulation
	 *    wrap up and clean up
	 *    write final results
	 */
	public void end() throws Throwable
	{
		double totalSize = 0.0;
        int matchingVisits = 0;
        int matchingVisitsWithConfidence = 0;
        int totalPossibleIntervalVisits = 0;
        int totalPossibleConfidenceScore = 0;
        int totalVisits = 0;
        int totalVisitsOutside = 0;
        int totalMatchedIntervals = 0;
        double totalMatchedIntervalScore = 0.0;
        int numberOfCheckpoints = 0;
        final ArrayList<Checkpoint> cstList = getCheckpointList();
        final ArrayList<HomininFactory> factoryList = getFactoryList();
		final ArrayList<HomininGroup> groupList = getHomininList();		
        int totalVisitsAsHomerangeInsideInterval = 0;
        int totalVisitsAsHomerangeOutsideInterval = 0;
        int totalVisitsAsHomerangeInsideIntervalWithConfidence = 0;
		
        simulationHasEnded = true;
        
		logger.info("Ending simulation");
        logger.info("");
		
		// final topographic display when the user interface is active
	    for (final HabitatCell habitat : habitatList) 
	    {
	    	// Select what should be shown; also set this in HSUtils!
	    	//habitat.showNumberOfDeads();
        	//habitat.showNumberOfVisits();
        	habitat.ShowValue();
	    }
		
		deadHomininList.clear();
		
		// collect results; calculate totals
        for (final Checkpoint cst : cstList) 
        {
        	if (cst.getType() == Constants.CheckpointType.CHECKPOINT_TYPE_STARTINGPOINT_CST)  
        	{
        		logger.info("Starting checkpoint " + cst + " visited (does NOT count towards score): " + cst.getVisitsOutsideIntervals());
        	}
        	else
          	if (cst.getType() == Constants.CheckpointType.CHECKPOINT_TYPE_CLIMATE_CST)  
           	{
           		logger.info("Climate monitoring checkpoint " + cst + " visited (does NOT count towards score): " + cst.getVisitsOutsideIntervals());
           	}
           	else
            	if (cst.getType() == Constants.CheckpointType.CHECKPOINT_TYPE_MONITORING_CST)
            	{
            		logger.info("Monitoring checkpoint " + cst + " visited (does NOT count towards score): " + cst.getVisitsOutsideIntervals() + ". \t\tPresence in: " + cst.getMISResults());
            	}
            	else
            	{
            		logger.info("Checkpoint " + cst + " #matches: " + cst.getVisitsInsideIntervals() + "("+ HSUtils.getPercentageString(cst.getMatchingPercentage(), 100.0) + "). Visited outside intervals: " + cst.getVisitsOutsideIntervals());
            		numberOfCheckpoints++;
            		matchingVisits += cst.getVisitsInsideIntervals();
            		matchingVisitsWithConfidence += cst.getVisitsInsideIntervalsWithConfidence();
            		totalVisits += cst.getYearsWithVisits();
            		totalVisitsOutside += cst.getVisitsOutsideIntervals();
            		totalPossibleIntervalVisits += cst.getTotalIntervalSize();
            		totalMatchedIntervals += cst.getMatchedIntervals();
            		totalMatchedIntervalScore += cst.getMatchedIntervals() + cst.getMatchedIntervalCoverage();
            		totalPossibleConfidenceScore += cst.getTotalIntervalConfidenceScore();
                    totalVisitsAsHomerangeInsideInterval += cst.getVisitsAsHomerangeInsideIntervals();
                    totalVisitsAsHomerangeOutsideInterval += cst.getVisitsAsHomerangeOutsideIntervals();
                    totalVisitsAsHomerangeInsideIntervalWithConfidence += cst.getVisitsAsHomerangeInsideIntervalsWithConfidence();
            		
            	}
        	cst.printResults();
        }
        
		double type0 = 0, type1 = 0, type2 = 0;
		int groupType0 = 0, groupType1 = 0, groupType2 = 0;
        for (final HomininGroup group : groupList) 
        {
        	double size = group.getSize();
        	totalSize += size;
        	if (group.getHomininType() == 0) 
        	{
        		type0 += size;
        		groupType0++;
        	}
        	if (group.getHomininType() == 1) 
        	{
        		type1 += size;
    			groupType1++;
        	}
        	if (group.getHomininType() == 2)
        	{
        		type2 += size;
    			groupType2++;
        	}
        	
        	// remove groups from view
        	theContext.remove(group);
        }

        double totalPercentages = 0.0;

        // Final summary per checkpoint
        logger.info("");
        for (final Checkpoint cst : cstList) 
        {
        	if (cst.getType() == Constants.CheckpointType.CHECKPOINT_TYPE_GENERAL_CST)
        	{
        		logger.info(String.format("%-36s", cst) + " visits: " + String.format("%7d", cst.getYearsWithVisits()) + "\t\t matches: " + String.format("%7d", cst.getVisitsInsideIntervals()) + " (" + HSUtils.getPercentageString(cst.getMatchingPercentage(), 100) + ").  \tPresence in: " + cst.getMISResults() + "   homerange: " + cst.getVisitsAsHomerangeInsideIntervals() + "; " + cst.getVisitsAsHomerangeOutsideIntervals() + "; " + cst.getVisitsAsHomerangeInsideIntervalsWithConfidence());
        		totalPercentages += cst.getMatchingPercentage();
        	}
        }
        
        logger.info("");
        
		logParameters();
		
		// Administrativia
		long t = System.currentTimeMillis();
		logger.info("Finished simulation in " + Math.round((t-startTime)/1000) + " seconds  (" + Math.round((t-startTime)/60000) + " minutes)."); //  (" + (t-startTime) + " ms.)");

		for (final HomininFactory factory : factoryList) 
        {
        	logger.info("Factory " + factory + " produced: \t" + factory.getCreatedGroups() + " groups." );
        }
		logger.info("New groups created as offspring:             " + Math.round(HSUtils.getHomininOffspring()) );
        logger.info("");

        logger.info("Number of groups present: \t" + groupList.size() + ", per type: (" + groupType0 + ", "+ groupType1 + ", " + groupType2 + ")");
        logger.info("Number of hominins present: \t" + Math.round(totalSize) + ", per type: (" + Math.round(type0) + ", "+ Math.round(type1) + ", " + Math.round(type2) + ")");
		int sum = HSUtils.getHomininGroupDeathCount(0) + HSUtils.getHomininGroupDeathCount(1) + HSUtils.getHomininGroupDeathCount(2);
		logger.info("Number of hominin groups removed: " + sum +  
				" (" + HSUtils.getHomininGroupDeathCount(0) + ", " + HSUtils.getHomininGroupDeathCount(1) + ", "+ HSUtils.getHomininGroupDeathCount(2) + ").");
		logger.info("Number of hominins that died: " + String.format("%,14d", (Math.round(HSUtils.getHomininDeathCount(0)) + Math.round(HSUtils.getHomininDeathCount(1)) + Math.round(HSUtils.getHomininDeathCount(2)))) + 
				" (" + String.format("%,14d", Math.round(HSUtils.getHomininDeathCount(0))) + ", " + String.format("%,14d", Math.round(HSUtils.getHomininDeathCount(1))) + ", " + 
				String.format("%,14d", Math.round(HSUtils.getHomininDeathCount(2))) + ")." );

		logger.info("Groups that died crossing the border:        " + Math.round(HSUtils.getHomininGroupDeathCountPerCause(Constants.CAUSE_OF_DEATH_BORDER)) );
		logger.info("Groups that died crossing into Iberia:       " + Math.round(HSUtils.getHomininGroupDeathCountPerCause(Constants.CAUSE_OF_DEATH_BORDER_IBERIA)) );
		logger.info("Groups that died from extreme cold:          " + Math.round(HSUtils.getHomininGroupDeathCountPerCause(Constants.CAUSE_OF_DEATH_COLD)) );
		logger.info("Groups that died on flooded land:            " + Math.round(HSUtils.getHomininGroupDeathCountPerCause(Constants.CAUSE_OF_DEATH_FLOODED)) );
		logger.info("Groups that died from hunger:                " + Math.round(HSUtils.getHomininGroupDeathCountPerCause(Constants.CAUSE_OF_DEATH_HUNGER)) );
		logger.info("Groups that merged with other groups:        " + Math.round(HSUtils.getHomininGroupDeathCountPerCause(Constants.CAUSE_OF_DEATH_MERGER)) );
		logger.info("Groups that died because they are too small: " + Math.round(HSUtils.getHomininGroupDeathCountPerCause(Constants.CAUSE_OF_DEATH_SIZE)) );
        logger.info("");

        logger.info("All possible checkpoint visits:                    " + String.format("%,14d", numberOfCheckpoints * tickCount));
        logger.info("All realized checkpoint visits:                    " + String.format("%,14d", totalVisits) + "   (" + HSUtils.getPercentageString(totalVisits, numberOfCheckpoints * tickCount) + ")");
        logger.info("All visits outside any interval:                   " + String.format("%,14d", totalVisitsOutside) );
        logger.info("All visits outside any interval as homebase:       " + String.format("%,14d", totalVisitsAsHomerangeOutsideInterval) );
        logger.info("All possible interval visits:                      " + String.format("%,14d", totalPossibleIntervalVisits));
        logger.info("All realized interval visits:                      " + String.format("%,14d", matchingVisits) + "   (" + HSUtils.getPercentageString(matchingVisits, totalPossibleIntervalVisits) + ")");
        logger.info("All realized interval visits as homebase:          " + String.format("%,14d", totalVisitsAsHomerangeInsideInterval) + "   (" + HSUtils.getPercentageString(totalVisitsAsHomerangeInsideInterval, totalPossibleIntervalVisits) + ")");
        
        logger.info("Maximum possible confidence score interval visits: " + String.format("%,14d", totalPossibleConfidenceScore) );
        logger.info("Realized confidence score for all interval visits: " + String.format("%,14d", matchingVisitsWithConfidence) + "   (" + HSUtils.getPercentageString(matchingVisitsWithConfidence, totalPossibleConfidenceScore) + ")");
        logger.info("              as homebase for all interval visits: " + String.format("%,14d", totalVisitsAsHomerangeInsideIntervalWithConfidence) + "   (" + HSUtils.getPercentageString(totalVisitsAsHomerangeInsideIntervalWithConfidence, totalPossibleConfidenceScore) + ")");

        logger.info("");
        logger.info("Sum all checkpoint matching percentages:           " + Math.round(totalPercentages*100)/100 + "   (" + numberOfCheckpoints + " checkpoints)");
        logger.info("  Average checkpoint matching percentage:          " + Math.round((totalPercentages / numberOfCheckpoints) * 100) / 100 + "%");

		// write output to global results file
		try {
			
			File file = new File("HS Results " + simulationId + " " + timeStampName + ".txt");
			 
			// if file does not exists, then create it
			if (!file.exists()) {
				file.createNewFile();
			}
	
			FileWriter fw = new FileWriter(file.getAbsoluteFile());
			BufferedWriter bw = new BufferedWriter(fw);
			
			//
    		// header
			//
			String content = parameterHeader.replace("\n", "").replace("\r", "");
			bw.write(content);

			content =  "," +
					"use_energy" + "," + 
					"use_static" + "," + 
					"zero_water" + "," + 
					"zero_energy" + "," + 
					"zero_randomwalk" + "," + 
					"created_groups" + "," + 
					"created_hominins" + "," + 
					"totalVisits" + "," + 
					"matchingVisits" + "," + 
					"matchedIntervals" + "," + 
					"matchedIntervalCoverage" + "," +
					"totalVisitsAsHomerangeInsideInterval" + "," + 
					"totalVisitsAsHomerangeOutsideInterval" +"," + 
					"matchingVisitsWithConfidence";
			bw.write(content);

	        // include all factories and new groups created as offspring from existing groups
			for (final HomininFactory factory : factoryList) 
	        {
				content = "," + factory.getName();
				bw.write(content);
	        }
			content = "," + "Offspring";
			bw.write(content);

			// results for all checkpoints, all types
	        for (final Checkpoint cst : cstList) 
	        {
				content = "," +
						cst.getName()+"_totalVisits" + "," + 
						cst.getName()+"_matchingVisits" + "," + 
						cst.getName()+"_totalVisitsAsHomerangeInsideInterval" + "," + 
						cst.getName()+"_totalVisitsAsHomerangeOutsideInterval" +"," + 
						cst.getName()+"_matchedIntervals" +"," + 
						cst.getName()+"_matchedIntervalCoverage" +"," + 
						cst.getName()+"_matchingVisitsWithConfidence";
				bw.write(content);
	        }

	        bw.write(" ");
			bw.newLine();

			//
			// collect simulation result values
			//
			
    		// start with input parameters
			content = parameterLine;
			if (content != null)
			{
				// remove eol
				content = content.replace("\n", "");
				content = content.replace("\r", "");
			}
			bw.write(content);

			// constant settings for this simulation
			content = "," + 
					Constants.ENERGY_CONTINUOUS + "," + 
					Constants.STATIC_DISPERSAL + "," + 
					Constants.ZERO_MODEL_DISABLE_ENERGY + "," + 
					Constants.ZERO_MODEL_DISABLE_WATER + "," + 
					Constants.ZERO_MODEL_RANDOM_WALK; 
			bw.write(content);
			
			// output collected statistical results
			int groupCount = groupList.size() + HSUtils.getHomininGroupDeathCount(0) + HSUtils.getHomininGroupDeathCount(1) + HSUtils.getHomininGroupDeathCount(2);
			long homininCount = Math.round(totalSize) + Math.round(HSUtils.getHomininDeathCount(0)) + Math.round(HSUtils.getHomininDeathCount(1)) + Math.round(HSUtils.getHomininDeathCount(2));
			content = "," + 
					groupCount + "," + 
					homininCount + "," + 
					totalVisits + "," + 
					matchingVisits + "," + 
					totalMatchedIntervals + "," +
					Math.round(totalMatchedIntervalScore*100) + "," +
					totalVisitsAsHomerangeInsideInterval + "," + 
					totalVisitsAsHomerangeOutsideInterval +"," + 
					matchingVisitsWithConfidence;
			bw.write(content);

	        // results for all factories
			for (final HomininFactory factory : factoryList) 
	        {
				content = "," + factory.getCreatedGroups();
				bw.write(content);
	        }
			
			// number of offspring
			content = "," + Math.round(HSUtils.getHomininOffspring());
			bw.write(content);
	        
			// results for all checkpoints
	        for (final Checkpoint cst : cstList) 
	        {
	        	double matchedScore = cst.getMatchedIntervals() + cst.getMatchedIntervalCoverage();
				content = "," +
						cst.getYearsWithVisits() + "," + 
						cst.getVisitsInsideIntervals() + "," + 
						cst.getVisitsAsHomerangeInsideIntervals() + "," + 
						cst.getVisitsAsHomerangeOutsideIntervals() +"," + 
						cst.getMatchedIntervals() +"," + 
						Math.round(matchedScore*100) +"," + 
						cst.getVisitsAsHomerangeInsideIntervalsWithConfidence();
				bw.write(content);
	        }

			bw.newLine();
			bw.flush();
			bw.close();

		} catch (IOException e) {
			e.printStackTrace();
			logger.info("Error when writing results" );
		}

		// log all climate checkpoint output
	    for (final Checkpoint checkpoint : climatecheckpointList) 
	    {
    		checkpoint.writeClimateData(simulationId + " " + timeStampName);
	    }

	    // cleanup
	    climatecheckpointList.clear();
		checkpointList.clear();
        for (final Checkpoint cst : cstList) 
        {
        	cst.finalize();
        }
        
		// release memory (sub contexts if needed)
		System.gc();
		
		logger.info("Simulation ended" );
		logFileHandler.flush();
		logFileHandler.close();
	}	// end
	
	/**
	 * Implementation of the agent activity in each turn.
	 *   ScheduledMethod Does NOT work
	 */
	//@ScheduledMethod(start = 1, interval = 1, priority = 10)
	public static void activateAgents() 
	{
		double tick = getTickCount();
        double totalGroupCount = 0, hungryGroupCount = 0;
        double totalPopulation = 0, hungryPopulation = 0;
		double sizeType0 = 0, sizeType1 = 0, sizeType2 = 0;
		int groupCountType0 = 0, groupCountType1 = 0, groupCountType2 = 0;
		boolean hasGroupMerged = false;
        
		// get globals for this turn
		globalSealevel = HSUtils.getSeaLevel((int)Math.round(tick));
		globalTemperatureOffset = HSUtils.getTemperature((int)Math.round(tick));
		//Constants.DEBUG_PRINT_LEVEL("Tickcount: " + Math.round(tick) + " sealevel: " + getGlobalSeaLevel() + " temp: " + getGlobalTemperature(), 30);

		// update HabitatCells with climate and production for this year
	    for (final HabitatCell habitat : habitatList) 
	    {
        	habitat.updateHabitatCell();
	    }

	    // do HomininGroups, in randomized order
		final ArrayList<HomininGroup> groupList = getHomininList();
        SimUtilities.shuffle(groupList, RandomHelper.getUniform());

        // move  the hominin groups 
        for (final HomininGroup group : groupList) 
        {
        	if (group.isAlive())
        	{
        		group.move();
        	}
        }
        
        // consume energy from the landscape
        for (final HomininGroup group : groupList) 
        {
        	if (group.isAlive())
        	{
        		group.eat();
        	}
        }
	     
        // demographics
        for (final HomininGroup group : groupList) 
        {
        	if (group.isAlive())
        	{
	        	group.grow();
	        	
	        	if (group.getAge() > yearsBeforeGroupMaturity)
	        	{
	        		// only allow mature groups to merge when too small
	               	if (group.isSmall())
	            	{
	            		// try to merge with other groups in the area
	            		hasGroupMerged = group.merge();
		        	}
	        	}
	        	
	        	if (!hasGroupMerged)
	        	{
	        		// reproduction
	        	    group.reproduce();
	        	}
        	}
        	
        	// administration
        	if (group.checkAlive())
        	{
        		// group is still alive, update statistics
            	double size = group.getSize();
            	
        		totalGroupCount++;
        		totalPopulation += size;
        		
        		if (group.isHungry())
        		{
        			hungryGroupCount++;
        			hungryPopulation += size;
        		}
            	if (group.getHomininType() == 0) 
            	{
            		sizeType0 += size;
            		groupCountType0++;
            	} else
            	if (group.getHomininType() == 1) 
            	{
            		sizeType1 += size;
        			groupCountType1++;
            	} else
            	if (group.getHomininType() == 2)
            	{
            		sizeType2 += size;
        			groupCountType2++;
            	}
        	}
        }
        
        hungryGroupPercentage = Math.round((hungryGroupCount / totalGroupCount) * 100);
        hungryHomininPercentage = Math.round((hungryPopulation / totalPopulation) * 100);
        
        // every 100 turns, provide debug info
		if (getTickCount() /100 == Math.round(getTickCount()/100))
		{
			Constants.DEBUG_PRINT_LEVEL("Number of removed groups: (" + HSUtils.getHomininGroupDeathCount(0)+"; "+ HSUtils.getHomininGroupDeathCount(1)+"; "+ HSUtils.getHomininGroupDeathCount(2)+")" + 
					". Groups present: " + totalGroupCount + " (" + groupCountType0 + "; " + groupCountType1 + "; " + groupCountType2 + 
					"). Hominins present: " + Math.round(totalPopulation*100)/100 + " ( " + Math.round(sizeType0*100)/100 + "; " + Math.round(sizeType1*100)/100 + "; " + Math.round(sizeType2*100)/100 + ").", 4);
				
			//System.gc();   // explicit garbage control, but does not make a performance difference, so disabled for now
		}
			
        /*
         * Check for production with the factories after groups have eaten
         */
		if (useFactories) 
		{
	        for (final HomininFactory factory : factoryList) 
	        {
       		   	factory.produce(false);
	        }
		}
		
		/*
		 * Update climate monitoring checkpoints
		 */
    	int currentYear = HomininSpaceContextBuilder.getYear();
	    for (final Checkpoint checkpoint : climatecheckpointList) 
	    {
    		checkpoint.monitorClimate(currentYear);
	    }
	        
	}	// activateAgents

	
/*
 * Current day Sea cells do not have an temperature nor precipitation, extend from land
 * 		This assumes that current day climate extends over sea (model)
 *		LGM has lowest sea level for all sea cells, and thus temp and precip for each cell that will be used
 */
private void ExtendLandClimate()
{
    // For those locations on the grid that are sea level, create artificial temperature, precipitation
    // 		first, for each latitude line (y) go from right to left, then from left to right
	//		then vertical extrapolation for those cells that were missed in the horizontal extrapolation
    double landTemperature = 0.0;
    double landWarmestTemperature = 0.0;
    double landColdestTemperature = 0.0;
    double landPrecipitation = 0.0;
    boolean fromLand;
    
    // first horizontal extrapolation
    for (int y=0; y<gridSizeY; y++)
    {
    	// from right to left
    	fromLand = false;
    	for (int x=gridSizeX-1; x>=0; x--)
    	{
            final Iterable<Object> objects = theGrid.getObjectsAt(x,y);

            for (final Object object : objects) 
            {
            	if (object instanceof HabitatCell) 
            	{
            		HabitatCell h = (HabitatCell) object;
            		if ((h.getHabitatType() == Constants.HABITAT_TYPE_SEA) || (h.getHabitatType() == Constants.HABITAT_TYPE_LAKE) || (h.getCurrentPrecipitation()<0.01))
            		{
            			if (fromLand)
            			{
                			h.setCurrentTemperature(landTemperature);
                			h.setCurrentTemperatureWarmestMonth(landWarmestTemperature);
                			h.setCurrentTemperatureColdestMonth(landColdestTemperature);
                			h.setCurrentPrecipitation(landPrecipitation);
            			}
            		}
            		else
            		{
            			landTemperature = h.getCurrentTemperature();
            			landWarmestTemperature = h.getCurrentWarmestMonthTemperature();
            			landColdestTemperature = h.getCurrentColdestMonthTemperature();
            			landPrecipitation = h.getCurrentPrecipitation();
            			fromLand = true;
            		}
             	}
            }
      	}
    	
    	// from left to right
    	fromLand = false;
    	for (int x=0; x<gridSizeX; x++)
    	{
            final Iterable<Object> objects = theGrid.getObjectsAt(x,y);

            for (final Object object : objects) {
            	if (object instanceof HabitatCell) {
            		HabitatCell h = (HabitatCell) object;
            		if ((h.getHabitatType() == Constants.HABITAT_TYPE_SEA) || (h.getHabitatType() == Constants.HABITAT_TYPE_LAKE) || (h.getCurrentPrecipitation()<0.01))
            		{
            			if (fromLand)
            			{
            				if (h.getCurrentTemperature() < landTemperature)
            				{
            				   	// new temperature
	                				h.setCurrentTemperature(landTemperature);
		                			h.setCurrentTemperatureWarmestMonth(landWarmestTemperature);
		                			h.setCurrentTemperatureColdestMonth(landColdestTemperature);
	                				h.setCurrentPrecipitation(landPrecipitation);
            				}
            			}
            		}
            		else
            		{
            			fromLand = true;
            			landTemperature = h.getCurrentTemperature();
            			landWarmestTemperature = h.getCurrentWarmestMonthTemperature();
            			landColdestTemperature = h.getCurrentColdestMonthTemperature();
            			landPrecipitation = h.getCurrentPrecipitation();
            		}
             	}
            }
      	}
    }

    // now vertical extrapolation
    for (int x=0; x<gridSizeX; x++)
    {
    	// from bottom to top
    	fromLand = false;
    	for (int y=0; y<gridSizeY; y++)
    	{
            final Iterable<Object> objects = theGrid.getObjectsAt(x,y);

            for (final Object object : objects) {
            	if (object instanceof HabitatCell) {
            		HabitatCell h = (HabitatCell) object;
            		if ((h.getHabitatType() == Constants.HABITAT_TYPE_SEA) || (h.getHabitatType() == Constants.HABITAT_TYPE_LAKE))
            		{
            			if (fromLand)
            			{
            				if (h.getCurrentTemperature() < landTemperature)
            				{
            				   	// new temperature
	                				h.setCurrentTemperature(landTemperature);
		                			h.setCurrentTemperatureWarmestMonth(landWarmestTemperature);
		                			h.setCurrentTemperatureColdestMonth(landColdestTemperature);
	                				h.setCurrentPrecipitation(landPrecipitation);
            				}
            			}
            		}
            		else
            		{
            			fromLand = true;
            			landTemperature = h.getCurrentTemperature();
            			landWarmestTemperature = h.getCurrentWarmestMonthTemperature();
            			landColdestTemperature = h.getCurrentColdestMonthTemperature();
            			landPrecipitation = h.getCurrentPrecipitation();
            		}
             	}
            }
      	}
    	
    	// from top to bottom
    	fromLand = false;
    	for (int y=gridSizeY-1; y>=0; y--)
    	{
            final Iterable<Object> objects = theGrid.getObjectsAt(x,y);

            for (final Object object : objects) {
            	if (object instanceof HabitatCell) {
            		HabitatCell h = (HabitatCell) object;
            		if ((h.getHabitatType() == Constants.HABITAT_TYPE_SEA) || (h.getHabitatType() == Constants.HABITAT_TYPE_LAKE))
            		{
            			if (fromLand)
            			{
            				if (h.getCurrentTemperature() < landTemperature)
            				{
            				   	// new temperature
                				h.setCurrentTemperature(landTemperature);
	                			h.setCurrentTemperatureWarmestMonth(landWarmestTemperature);
	                			h.setCurrentTemperatureColdestMonth(landColdestTemperature);
                				h.setCurrentPrecipitation(landPrecipitation);
            				}
            			}
            		}
            		else
            		{
            			fromLand = true;
            			landTemperature = h.getCurrentTemperature();
            			landWarmestTemperature = h.getCurrentWarmestMonthTemperature();
            			landColdestTemperature = h.getCurrentColdestMonthTemperature();
            			landPrecipitation = h.getCurrentPrecipitation();
            		}
             	}
            }
      	}
    }
    
}	// ExtendLandClimate
        
	

	/*
	 * Create or take from list of dead groups new hominin group
	 * 		This list is maintained for performance reasons (creating new objects takes more time than re-using them)
	 */
	public static HomininGroup getNewHomininGroup()
	{
		HomininGroup newGroup = null;
	
		if (deadHomininList.size() > 0)
		{
			newGroup = deadHomininList.get(0);
			deadHomininList.remove(0);
		}
		else
		{
			newGroup = new HomininGroup(theGrid);
		}
		
		theContext.add(newGroup);
		newGroup.initialize( getNewAgentId() );
        newGroup.setSize(Constants.INITIAL_SIZE_PREFERTILE, Constants.INITIAL_SIZE_FERTILE, Constants.INITIAL_SIZE_POSTFERTILE);
		
		return newGroup;
		
	}	// getNewHomininGroup

	/*
	 * Delete group
	 */
	public static void deleteHomininGroup( HomininGroup group )
	{
		theContext.remove(group);
		
		group.setSize(0.0, 0.0, 0.0);

		deadHomininList.add(group);
		
	}	// deleteHomininGroup
	
	/**
	 * Returns the current active {@link HomininGroup} list.
	 * 
	 */
	private static ArrayList<HomininGroup> getHomininList() 
	{
        @SuppressWarnings("unchecked")
        final Iterable<HomininGroup> hominins = RunState.getInstance().getMasterContext()
                        .getObjects(HomininGroup.class);
        final ArrayList<HomininGroup> homininList = new ArrayList<HomininGroup>();

        for (final HomininGroup hominin : hominins) {
        	homininList.add(hominin);
        }

        return homininList;
        
	}	// getHomininList

	/**
	 * Returns the current active {@link HomininFactory} list.
	 */
	private static ArrayList<HomininFactory> getFactoryList() 
	{
        @SuppressWarnings("unchecked")
        final Iterable<HomininFactory> hominins = RunState.getInstance().getMasterContext()
                        .getObjects(HomininFactory.class);
        final ArrayList<HomininFactory> homininList = new ArrayList<HomininFactory>();

        for (final HomininFactory hominin : hominins) {
        	homininList.add(hominin);
        }

        return homininList;
        
	}	// getFactoryList

	/**
	 * Returns the current active {@link Checkpoint} list.
	 */
	private static ArrayList<Checkpoint> getCheckpointList() 
	{
        @SuppressWarnings("unchecked")
        final Iterable<Checkpoint> csts = RunState.getInstance().getMasterContext()
                        .getObjects(Checkpoint.class);
        final ArrayList<Checkpoint> cstList = new ArrayList<Checkpoint>();

        for (final Checkpoint cst : csts) {
        	cstList.add(cst);
        }

        return cstList;
        
	}	// getCheckpointList

	/**
	 * Returns the current active {@link Checkpoint} list.
	 */
	private static ArrayList<Checkpoint> getClimateCheckpointList() 
	{
        @SuppressWarnings("unchecked")
        final Iterable<Checkpoint> csts = RunState.getInstance().getMasterContext()
                        .getObjects(Checkpoint.class);
        final ArrayList<Checkpoint> cstList = new ArrayList<Checkpoint>();

        for (final Checkpoint cst : csts) 
        {
        	if (cst.getType() == Constants.CheckpointType.CHECKPOINT_TYPE_CLIMATE_CST)
        	{
        		cstList.add(cst);
        	}
        }

        return cstList;
        
	}	// getClimateCheckpointList

	/**
	 * Returns the current active {@link HabitatCell} list.
	 */
	private static ArrayList<HabitatCell> getHabitatCellList() 
	{
        @SuppressWarnings("unchecked")
        final Iterable<HabitatCell> habitats = RunState.getInstance().getMasterContext()
                        .getObjects(HabitatCell.class);
        final ArrayList<HabitatCell> hList = new ArrayList<HabitatCell>();

        for (final HabitatCell habitat : habitats) {
        	hList.add(habitat);
        }

        return hList;
        
	}	// getHabitatCellList
	
	/*
	 * Check to see if simulation has ended
	 */
	public static boolean getSimulationHasEnded()
	{
		if ((simulationHasEnded == false) && (Math.round(getTickCount()) >= HSUtils.numberOftimeSteps))
		{
			return true;
		}
		return simulationHasEnded;
		
	}	// getSimulationHasEnded
	
	public static double getTickCount()
	{
		ISchedule schedule = RunEnvironment.getInstance().getCurrentSchedule();
		return schedule.getTickCount();
	}	// getTickCount
	
	/*
	 * Get current years, in years before present
	 */
	public static int getYear()
	{
		int year = HSUtils.startingTimeStep - (int) getTickCount() + 1;
		return year;
	}	// getYear
	
	/*
	 * Get and set functions
	 */
	public static int getNewAgentId()
	{
		return agentID++;
	}
	public static int getGridSizeX()
	{
		return gridSizeX;
	}
	public static int getGridSizeY()
	{
		return gridSizeY;
	}
	public static double getGlobalSeaLevel()
	{
		return globalSealevel;
	}
	public static double getGlobalTemperatureOffset()
	{
		return globalTemperatureOffset;
	}
	public static double getHungryGroupPercentage()
	{
		return hungryGroupPercentage;
	}
	public static double getHungryHomininPercentage()
	{
		return hungryHomininPercentage;
	}
	public static double getMinimumTemperature()
	{
		return minimumRequiredTemperature;
	}
	public static int getBirthsPercentage()
	{
		return birthsPercentage;
	}
	public static boolean getOpenBorders()
	{
		return openBorders;
	}
	public static boolean getStaticDispersal()
	{
		return staticDispersal;
	}
	public static int getSizePreFertileGroup()
	{
		return sizePreFertileGroup;
	}
	public static int getSizeFertileGroup()
	{
		return sizeFertileGroup;
	}
	public static int getYearsBeforeGroupMaturity()
	{
		return yearsBeforeGroupMaturity;
	}
	public static int getDeathsPreFertilePercentage()
	{
		return deathsPreFertilePercentage;
	}
	public static int getDeathsFertilePercentage()
	{
		return deathsFertilePercentage;
	}
	public static int getDeathsPostFertilePercentage()
	{
		return deathsPostFertilePercentage;
	}
	public static int getConsumptionPreFertile()
	{
		return consumptionPreFertile;
	}
	public static int getConsumptionFertile()
	{
		return consumptionFertile;
	}
	public static int getConsumptionPostFertile()
	{
		return consumptionPostFertile;
	}
	public static int getCaloriesPerKgMeat()
	{
		return caloriesPerKgMeat;
	}
	public static int getMaxForagingRange()
	{
		if (Constants.USE_MAXIMUM_FORAGING_RANGE)
			return maxForagingRange;
		else
			return gridSizeX;  // virtually unlimited foraging range
	}
	public static int getGroupSizeLarge()
	{
		return groupSizeLarge;
	}
	public static int getGroupSizeSmall()
	{
		return groupSizeSmall;
	}
	public static int getGroupSizeFertileSmall()
	{
		return groupSizeFertileSmall;
	}
	public static int getPrecipitationColdArid()
	{
		return precipitationColdArid;
	}
	public static int getPrecipitationWarmArid()
	{
		return precipitationWarmArid;
	}
	public static int getBiomassRecoverPercentage()
	{
		return biomassRecoverPercentage;
	}
	public static int getSimulationId()
	{
		return simulationId;
	}
}