/*
 Copyright 2011-2018 Fulco Scherjon

This file is part of the HomininSpace modelling and simulation system.

HomininSpace is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, any version 3.
HomininSpace is distributed WITHOUT ANY WARRANTY under the GNU General Public License.
HomininSpace comes with a copy of this GNU General Public License.  If not, see <http://www.gnu.org/licenses/>.

*/


package HomininSpace.agents;
import HomininSpace.HomininSpaceContextBuilder;
import HomininSpace.common.Constants;
import HomininSpace.common.HSUtils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import repast.simphony.space.grid.Grid;


/*
 * Confidence levels should be interpreted as follows:
 * 1.	unsure, data missing, disputed in the literature, only available through secondary sources;
 * 2.	reasonable, well dated element but with incomplete description or it is disputed;
 * 3.	very confident, well described, undisputed, complete results in literature.
 */


class ClimateDataPoint
{
	public int						year = 0;
	public double					temperature = -10000;
	public double					precipitation = -10000;
	public int						habitatType = 0;
	public int						climateType = 0;
}


class PresenceInterval
{
	public int 						earliestYear;		// all years in ky BP
	public int 						latestYear;
	private int						lastYear = -1;
	
	public int 						visitCount = 0;
	public int						visitCountPerType0 = 0;
	public int						visitCountPerType1 = 0;
	public int						visitCountPerType2 = 0;
	
	public boolean 					active = true;    // becomes false when all checked timepoints < end  ; to maximize performance
	public int						confidenceLevel = 1;  
	
	public PresenceInterval( int start, int end, int aConfidenceLevel )
	{
		earliestYear = start;
		latestYear = end;
		confidenceLevel = aConfidenceLevel;
		
		if (earliestYear < latestYear)
		{
			throw new IllegalStateException("Constructing presence interval: start after end.");
		}
	}
	
	/*
	 * Check if this point in time is within this interval
	 *   Note that if we are past the latestYear then this interval is turned off
	 */
	public boolean isInInterval( int pointInTime )
	{
		boolean res = false;

		if (active)
		{
			if ((pointInTime <= earliestYear) && (pointInTime >= latestYear))
			{
				res = true;
			}
			else
			{
				if (pointInTime < latestYear)
				{
					active = false;
				}
			}
		}
		
		return res;
	}
	
	/*
	 * This routine assumes it has already been tested that this point in time matches the interval
	 * 		Also note that visitCount can be less than the individual visits together, if two or more hominins are in the area at the same time
	 */
	public void setPresence( int thisYear, int typeHominin)
	{
		if (thisYear != lastYear)
		{
			visitCount++;
			lastYear = thisYear;
		}
		
		/* one visit per year, but we allow multiple visits per hominin type */
		if (typeHominin == 0)
			visitCountPerType0++;
		else if (typeHominin == 1)
			visitCountPerType1++;
		else if (typeHominin == 2)
			visitCountPerType2++;
		else 
			Constants.DEBUG_PRINT_LEVEL("Error in type Hominin in Interval: " + typeHominin, 5 );
	}

	public int getCoveragePercentage()
	{
		int res = 0;
		if (visitCount != 0)
			res = (int) Math.round(((double) getVisitCount() / (double) getSize()) * 100);
		return res;
	}
	
	public int getSize()
	{
		return earliestYear-latestYear+1;
	}
	public int getVisitCount()
	{
		return visitCount;
	}
	public int getVisitCountWithConfidence()
	{
		return visitCount * (confidenceLevel*confidenceLevel);
		//return visitCount * confidenceLevel;
	}
	
    @Override
    public String toString() {
            return String.format("Interval[%d,%d]: visits=%d (%d%%) (%d;%d;%d); ", earliestYear, latestYear, visitCount, getCoveragePercentage(), visitCountPerType0, visitCountPerType1, visitCountPerType2);
    }
}


/*
 * Checkpoint - Checkpoint in space and time
 * 		Agent that keep score of the presence near it of other mobile agents (HomininGroup)
 * 		Also climate monitoring for designated individuals at each time step
 */
public class Checkpoint {

	private int 						id = -1;          // unique identifier
	private String 						name;     // readable identification
	private int							posX = -1;
	private int							posY = -1;
	private Constants.CheckpointType	type;   // 1 = cst with intervals, 2 = presence monitoring point, 3 = starting point, 4 = climate monitor
	private int							lastYearWithPresence = -1;

	private Grid<Object> 				grid = null;
	private HabitatCell 				myHabitatCell 		= null;
	
	private boolean						isVisited = false;
	public boolean 						presenceMIS6 = false;    // becomes true when a visit in this period occurs
	public boolean 						presenceMIS5e = false;    // becomes true when a visit in this period occurs
	public boolean 						presenceMIS5d = false;    // becomes true when a visit in this period occurs
	public boolean 						presenceMIS5c = false;    // becomes true when a visit in this period occurs
	public boolean 						presenceMIS5b = false;    // becomes true when a visit in this period occurs
	public boolean 						presenceMIS5a = false;    // becomes true when a visit in this period occurs
	public boolean 						presenceMIS4 = false;    // becomes true when a visit in this period occurs
	public boolean 						presenceMIS3 = false;    // becomes true when a visit in this period occurs
	
	public int	 						presenceCountMIS6 = 0;
	public int	 						presenceCountMIS5e = 0;
	public int	 						presenceCountMIS5d = 0;
	public int	 						presenceCountMIS5c = 0;
	public int	 						presenceCountMIS5b = 0;
	public int	 						presenceCountMIS5a = 0;
	public int	 						presenceCountMIS4 = 0; 
	public int	 						presenceCountMIS3 = 0; 
	
	private int							yearsWithVisits = 0;
	
	private int							visitsOutsideIntervals = 0;
	private int							visitsInsideIntervals = 0;
	private int							visitsInsideIntervalsWithConfidence = 0;
	
	private int							visitsAsHomerangeOutsideIntervals = 0;
	private int							visitsAsHomerangeInsideIntervals = 0;
	private int							visitsAsHomerangeInsideIntervalsWithConfidence = 0;

	private ArrayList<PresenceInterval>	presenceIntervals;
	private ArrayList<ClimateDataPoint>	climateDataPoints;
	
	// Constructor
	public Checkpoint( int id, Constants.CheckpointType type )
	{
		this.id = id;
		this.name = "CST " + id;
		this.type = type; 
		
		presenceIntervals = new ArrayList<PresenceInterval>();
		climateDataPoints = new ArrayList<ClimateDataPoint>();
	}

	/*
	 * Finalize
	 */
	public void finalize() throws Throwable
	{
		presenceIntervals.clear();
		climateDataPoints.clear();
		
		super.finalize();
	} 
	
	// This function is called after all checkpoints have been read from file
	public void initialize( Grid<Object> grid )
	{
		this.grid = grid;
		
		if (null == grid) {
			throw new IllegalStateException("Checkpoint: Cannot locate grid in context.");
		}
		
		if (presenceIntervals.size() == 0)
		{
			Constants.DEBUG_PRINT_LEVEL("Checkpoint warning: No intervals defined for " + this.name , 50 );
		}
	}  // initialize

	// Extra function, to be called each time after setpresence if this cell is in the homerange of a group
	public void setPresenceAsHomeRange()
	{
		int currentYear = HomininSpaceContextBuilder.getYear();
		boolean currentYearInScanningInterval = false;
		
		// does this presence matches the archaeology, if so update presence count
		for (int i=0; i<presenceIntervals.size(); i++)
		{
			if (presenceIntervals.get(i).isInInterval(currentYear))
			{
				currentYearInScanningInterval = true;
				visitsAsHomerangeInsideIntervals++;
				visitsAsHomerangeInsideIntervalsWithConfidence += presenceIntervals.get(i).confidenceLevel*presenceIntervals.get(i).confidenceLevel;
				//visitsAsHomerangeInsideIntervalsWithConfidence += presenceIntervals.get(i).confidenceLevel;
			}
		}
			
		if (!currentYearInScanningInterval)
		{
			visitsAsHomerangeOutsideIntervals++;
		}
	}
	
	public void addHabitatCell( HabitatCell h)
	{
		myHabitatCell = h;

		if (null == myHabitatCell) {
			throw new IllegalStateException("Checkpoint: Cannot locate habitatcell in context." + this);
		}
	}
	
	/*
	 * monitorClimate
	 *    called once a year to register the current climate
	 */
	public void monitorClimate( int year )
	{
		ClimateDataPoint climate = new ClimateDataPoint();
		climate.year = year;
		climate.habitatType = myHabitatCell.getHabitatType();
		climate.climateType = myHabitatCell.getClimateType();
		climate.precipitation = myHabitatCell.getPrecipitation();
		climate.temperature = myHabitatCell.getTemperature();
		
		climateDataPoints.add(climate);
	}
	

	/*
	 * Indicate that an hominin group is in the area, and forages on this grid point
	 */
	public void setPresence( int id, int homininType )
	{
		int currentYear = HomininSpaceContextBuilder.getYear();
		
		if (lastYearWithPresence != currentYear)
		{
			boolean currentYearInScanningInterval = false;
			yearsWithVisits++;
			lastYearWithPresence = currentYear;
			isVisited = true;
			
			// contents of sets are always unique
			// does this presence matches the archaeology, if so update presence count
			// if there are overlapping intervals, visitsInsideIntervals can be greater than totalVisits
			for (int i=0; i<presenceIntervals.size(); i++)
			{
				if (presenceIntervals.get(i).isInInterval(currentYear))
				{
					currentYearInScanningInterval = true;
					presenceIntervals.get(i).setPresence(currentYear, homininType);
					visitsInsideIntervals++;
					visitsInsideIntervalsWithConfidence += presenceIntervals.get(i).confidenceLevel*presenceIntervals.get(i).confidenceLevel;
				}
			}
				
			if (!currentYearInScanningInterval)
			{
				visitsOutsideIntervals++;
			}
				
			{
				if ((currentYear > Constants.START_MIS5e))
				{
					presenceMIS6 = true;
					presenceCountMIS6++;
				}
			}
			{
				if ((currentYear <= Constants.START_MIS5e) && (currentYear > Constants.START_MIS5d))
				{
					presenceMIS5e = true;
					presenceCountMIS5e++;
				}
			}
			{
				if ((currentYear <= Constants.START_MIS5d) && (currentYear > Constants.START_MIS5c))
				{
					presenceMIS5d = true;
					presenceCountMIS5d++;
				}
			}
			{
				if ((currentYear <= Constants.START_MIS5c) && (currentYear > Constants.START_MIS5b))
				{
					presenceMIS5c = true;
					presenceCountMIS5c++;
				}
			}
			{
				if ((currentYear <= Constants.START_MIS5b) && (currentYear > Constants.START_MIS5a))
				{
					presenceMIS5b = true;
					presenceCountMIS5b++;
				}
			}
			{
				if ((currentYear <= Constants.START_MIS5a) && (currentYear > Constants.START_MIS4))
				{
					presenceMIS5a = true;
					presenceCountMIS5a++;
				}
			}
			{
				if ((currentYear <= Constants.START_MIS4) && (currentYear > Constants.START_MIS3))
				{
					presenceMIS4 = true;
					presenceCountMIS4++;
				}
			}
			{
				if ((currentYear <= Constants.START_MIS3) )
				{
					presenceMIS3 = true;
					presenceCountMIS3++;
				}
			}
		}
	}
	
	/*
	 * Write climate data; for climate monitoring points
	 *  
	 */
	public void writeClimateData( String id )
	{
		Constants.DEBUG_PRINT_LEVEL("Write climate data: " + this, 10 );
		Constants.DEBUG_PRINT_LEVEL("	HS Climate " + this.name + " " + id + ".txt", 5 );
		
		// write climate data to a results file
		try {
			
			File file = new File("HS Climate " + this + id + ".txt");
			 
			if (!file.exists()) {
				file.createNewFile();
			}
	
			FileWriter fw = new FileWriter(file.getAbsoluteFile());
			BufferedWriter bw = new BufferedWriter(fw);
			
    		// header
			String content = 
					"year" +  "," + 
					"temperature" + "," + 
					"precipitation" + "," + 
					"habitatType" + "," + 
					"climateType"; 
			bw.write(content);
	        bw.write(" ");
			bw.newLine();

			
	        for (final ClimateDataPoint p : climateDataPoints) 
	        {
				content = 
						p.year +  "," + 
						p.temperature + "," + 
						p.precipitation + "," + 
						p.habitatType + "," + 
						p.climateType; 
				bw.write(content);
		        bw.write(" ");
				bw.newLine();
	        }
	        
			bw.newLine();
			bw.flush();
			bw.close();
			//fw.flush();   // already closed

		} catch (IOException e) {
			e.printStackTrace();
		}

	}
	
	/*
	 * Print the results
	 */
	public void printResults()
	{
		String s = "";
		if (type == Constants.CheckpointType.CHECKPOINT_TYPE_GENERAL_CST)
		{
			Constants.DEBUG_PRINT_LEVEL("Visits within intervals: " + visitsInsideIntervals + ".   Outside: " + visitsOutsideIntervals, 5 );
			Constants.DEBUG_PRINT_LEVEL("Confidence score: " + visitsInsideIntervalsWithConfidence, 5 );
			Constants.DEBUG_PRINT_LEVEL("Interval results: " + presenceIntervals, 5 );
	
			int	[]					theVisitCountPerType;
			theVisitCountPerType = new int[Constants.TYPES_OF_HOMININS];

			theVisitCountPerType[0] = 0;
			for (int i=0; i<presenceIntervals.size(); i++)
			{
				theVisitCountPerType[0] += presenceIntervals.get(i).visitCountPerType0;
			}
			s += "0 (" + theVisitCountPerType[0] + ") ";
				
				theVisitCountPerType[1] = 0;
				for (int i=0; i<presenceIntervals.size(); i++)
				{
					theVisitCountPerType[1] += presenceIntervals.get(i).visitCountPerType1;
				}
				s += "1 (" + theVisitCountPerType[1] + ") ";
				
				theVisitCountPerType[2] = 0;
				for (int i=0; i<presenceIntervals.size(); i++)
				{
					theVisitCountPerType[2] += presenceIntervals.get(i).visitCountPerType2;
				}
				s += "2 (" + theVisitCountPerType[2] + ") ";
	        
			Constants.DEBUG_PRINT_LEVEL("Visits per type hominin: " + s, 5 );
		}
		
	} 
	
	public String getMISResults()
	{
		String s = "";
		if (presenceMIS6) s += "6 ("+presenceCountMIS6+" = " + HSUtils.getPercentageString(presenceCountMIS6, HSUtils.startingTimeStep-Constants.START_MIS5e ) + ")";
		if (presenceMIS5e) s += "5e ("+presenceCountMIS5e+" = " + HSUtils.getPercentageString(presenceCountMIS5e, Constants.START_MIS5e-Constants.START_MIS5d ) +")";
		if (presenceMIS5d) s += "5d ("+presenceCountMIS5d+" = " + HSUtils.getPercentageString(presenceCountMIS5d, Constants.START_MIS5d-Constants.START_MIS5c )+")";
		if (presenceMIS5c) s += "5c ("+presenceCountMIS5c+" = " + HSUtils.getPercentageString(presenceCountMIS5c, Constants.START_MIS5c-Constants.START_MIS5b )+")";
		if (presenceMIS5b) s += "5b ("+presenceCountMIS5b+" = " + HSUtils.getPercentageString(presenceCountMIS5b, Constants.START_MIS5b-Constants.START_MIS5a )+")";
		if (presenceMIS5a) s += "5a ("+presenceCountMIS5a+" = " + HSUtils.getPercentageString(presenceCountMIS5a, Constants.START_MIS5a-Constants.START_MIS4 )+")";
		if (presenceMIS4) s += "4 ("+presenceCountMIS4+" = " + HSUtils.getPercentageString(presenceCountMIS4, Constants.START_MIS4-Constants.START_MIS3 )+")";
		if (presenceMIS3) s += "3 ("+presenceCountMIS3+" = " + HSUtils.getPercentageString(presenceCountMIS3, Constants.START_MIS3-(HSUtils.startingTimeStep - HSUtils.numberOftimeSteps)+1 )+")";
		return s;
	}
	
	
	public int getId() 
	{
		return id;
	}
	public void setId(int id) 
	{
		this.id = id;
	}

	public String getName() 
	{
		return name;
	}
	public void setName(String name) 
	{
		this.name = name;
	}
	public Constants.CheckpointType getType()
	{
		return type;
	}
	
    public int getPosX() {
        return posX;
    }
    public int getPosY() {
        return posY;
    }
	public void setPosX(final int x) {
		this.posX = x;
	}
	public void setPosY(final int y) {
		this.posY = y;
	}
	public void setPosXY(final int x, final int y) {
		grid.moveTo(this, x, y);
		posX = x;
		posY = y;
	}

	public void addPresenceInterval(final int start, final int end, final int confidenceLevel) 
	{
		PresenceInterval interval = new PresenceInterval(start, end, confidenceLevel);
		presenceIntervals.add(interval);
	}
	
	/*
	 * check if this checkpoint has ever been visited by hominin groups
	 */
	public boolean visited()
	{
		return isVisited;
	}
	
	public int getVisitsOutsideIntervals()
	{
		return visitsOutsideIntervals;
	}
	public int getVisitsInsideIntervals()
	{
		return visitsInsideIntervals;
	}
	public int getVisitsInsideIntervalsWithConfidence()
	{
		return visitsInsideIntervalsWithConfidence;
	}
	public int getVisitsAsHomerangeOutsideIntervals()
	{
		return visitsAsHomerangeOutsideIntervals;
	}
	public int getVisitsAsHomerangeInsideIntervals()
	{
		return visitsAsHomerangeInsideIntervals;
	}
	public int getVisitsAsHomerangeInsideIntervalsWithConfidence()
	{
		return visitsAsHomerangeInsideIntervalsWithConfidence;
	}
	public int getYearsWithVisits()
	{
		return yearsWithVisits;
	}
	
	public double getMatchingPercentage()
	{
		double res = 0;
		double totalSize = 0;
		double totalCount = 0;
		for (int i=0; i<presenceIntervals.size(); i++)
		{
			totalSize += presenceIntervals.get(i).getSize();
			totalCount += presenceIntervals.get(i).getVisitCount();
		}
		if (totalSize > 0)
			res = (totalCount / totalSize) * 100;
		return res;
	}
	
	public double getAverageMatchingPercentage()
	{
		double res = 0;
		double count = 0;
		double totalSize = 0;
		double totalCount = 0;
		for (int i=0; i<presenceIntervals.size(); i++)
		{
			if (presenceIntervals.get(i).getVisitCount() > 0)
			{
				totalSize = presenceIntervals.get(i).getSize();
				totalCount = presenceIntervals.get(i).getVisitCount();
				res += (totalCount / totalSize) * 100;
				count++;
			}
		}
		if (count > 0)
			res = (res / count);
		return res;
	}
	
	public int getMatchedIntervals()
	{
		int res = 0;
		for (int i=0; i<presenceIntervals.size(); i++)
		{
			if (presenceIntervals.get(i).getVisitCount() > 0)
				res++;
		}
		return res;
	}
	
	// Return the sum of the coverage percentages divided by 100 (example: 30%+50%+60% = 1.10) 
	public double getMatchedIntervalCoverage()
	{
		double res = 0.0;
		for (int i=0; i<presenceIntervals.size(); i++)
		{
			if (presenceIntervals.get(i).getVisitCount() > 0)
				res += ((double) presenceIntervals.get(i).getCoveragePercentage() / 100.0); 
		}
		return res;
	}
	
	public double getTotalIntervalSize()
	{
		double totalSize = 0;
		for (int i=0; i<presenceIntervals.size(); i++)
		{
			totalSize += presenceIntervals.get(i).getSize();
		}
		return totalSize;
	}
	
	public double getTotalIntervalConfidenceScore()
	{
		double totalScore = 0;
		for (int i=0; i<presenceIntervals.size(); i++)
		{
			totalScore += presenceIntervals.get(i).getSize() * (presenceIntervals.get(i).confidenceLevel * presenceIntervals.get(i).confidenceLevel);
		}
		return totalScore;
	}
	
    @Override
    public String toString() {
            return String.format("%s [x=%d, y=%d]", this.name, posX, posY);
    }

}
