/*
 Copyright 2011-2018 Fulco Scherjon

This file is part of the HomininSpace modelling and simulation system.

HomininSpace is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, any version 3.
HomininSpace is distributed WITHOUT ANY WARRANTY under the GNU General Public License.
HomininSpace comes with a copy of this GNU General Public License.  If not, see <http://www.gnu.org/licenses/>.

*/

package HomininSpace.agents;

import HomininSpace.HomininSpaceContextBuilder;
import HomininSpace.common.CellData;
import HomininSpace.common.Constants;
import HomininSpace.common.HSUtils;

import java.util.ArrayList;
import java.util.List;

import repast.simphony.query.space.grid.GridCell;
import repast.simphony.util.ContextUtils;
import repast.simphony.valueLayer.GridValueLayer;

/*
 * HabitatCell - implements available energy for each grid cell
 */
public class HabitatCell {
    private final int 				x, y;												// location on the grid
    
    private boolean 				isHomeRange = false;
    private int						foragingGroups = 0;
    private int						foragingGroupsLastYear = 0;
    private int						foragingGroupsLastLastYear = 0;
    private ArrayList<Checkpoint>	theCheckpoints = null;

    private int 					habitatType = Constants.HABITAT_TYPE_SEA;			// default
    private int 					climateType = Constants.CLIMATE_TYPE_WARM_FOREST;	// default
    
    private double					secondaryBiomass;  									// in kg for this year
    private double					edibleMeat;  										// in kg for this year
    private double					eatenEdibleMeat;  									// in kg, this year
    private int						recoverPercentage = 0;								// percentage that recovers per year
    private int						remainingBiomassPercentage = 100;					// percentage that remained previous turn
    
    private double 					height = 0.0;   									// in meters relative to sea level today

    private int						numberOfVisits = 0;									// history, during all timesteps
    private int						numberOfDeads = 0;									// history, during all timesteps
    
    private int						precipitationWarmAridLevel = 0;						// when is climate warm and arid?
    private int						precipitationColdAridLevel = 0;						// when is climate cold and arid?
    
    private CellData 				cellData;
    private GridValueLayer 			visibleValueLayer = null;
    
	private List<GridCell<HabitatCell>> theNeighborhood = null;

    /*
     * Constructor with CellData object
     */
    public HabitatCell(final CellData cellData) 
    {
    	this.cellData = cellData;
    	x = cellData.getX();
    	y = cellData.getY();
    	
		if (x < 0) {
			throw new IllegalArgumentException(String.format(
					"Coordinate x = %d < 0.", x));
		}
		if (y < 0) {
			throw new IllegalArgumentException(String.format(
					"Coordinate y = %d < 0.", y));
		}
        
        habitatType = cellData.getType();    // per default take type from the celldata object
        setHeigth(cellData.getHeight(), false);
        
    }	// HabitatCell
    
    /*
     * Constructor without CellData object, this default starts as a sea grid cell
     */
    public HabitatCell(final int xPos, final int yPos) 
    {
    	x = xPos; y = yPos;
    	cellData = new CellData(x, y); 
    	cellData.setType(Constants.HABITAT_TYPE_SEA);
    	
		if (x < 0) {
			throw new IllegalArgumentException(String.format(
					"Coordinate x = %d < 0.", x));
		}
		if (y < 0) {
			throw new IllegalArgumentException(String.format(
					"Coordinate y = %d < 0.", y));
		}
        
        habitatType = Constants.HABITAT_TYPE_SEA;   // default
        setHeigth(0.0, false);
        
    }	// HabitatCell
    
    /*
     * General one time initialization, called just before simulation starts
     */
    public void initialize()
    {
        secondaryBiomass = 0.0;
        remainingBiomassPercentage = 100;
        edibleMeat = 0.0;
        eatenEdibleMeat = 0.0;
    	numberOfVisits = 0;
    	numberOfDeads = 0;
    	foragingGroups = foragingGroupsLastYear = foragingGroupsLastLastYear = 0;
    	setHeigth(cellData.getHeight(), false);
    	//updateHabitatCell();  not possible yet
    	
    	// fill in defaults from parameters
    	precipitationWarmAridLevel = HomininSpaceContextBuilder.getPrecipitationWarmArid();
    	precipitationColdAridLevel = HomininSpaceContextBuilder.getPrecipitationColdArid();
    	recoverPercentage = HomininSpaceContextBuilder.getBiomassRecoverPercentage();

    	visibleValueLayer = (GridValueLayer) ContextUtils
         		.getContext(this).getValueLayer(Constants.VISIBLE_VALUE_LAYER_ID);
    	
    }	// initialize
    
    /*
     * Update routine, called from main loop for each timestep
     */
    //@ScheduledMethod(start = 1, interval = 1, priority = 1)
    public void updateHabitatCell() 
    {
    	// new timestep
    	foragingGroupsLastLastYear = foragingGroupsLastYear;
    	foragingGroupsLastYear = foragingGroups;
    	foragingGroups = 0;
    	
        // Adjust currentHeight to current sea level
        double newHeight = cellData.getHeight();
        // with - the land sinks, with + the ocean rises
        newHeight += HomininSpaceContextBuilder.getGlobalSeaLevel();
        setHeigth(newHeight, false);
        
        if ((newHeight <= Constants.SEA_LEVEL_HEIGHT) && !Constants.ZERO_MODEL_DISABLE_WATER) 
        {
        	// this is under water, so no production
        	climateType = Constants.CLIMATE_TYPE_SEA;
        	secondaryBiomass = 0.0;
        	edibleMeat = 0.0;
        	eatenEdibleMeat = 0.0;
        }
        else
        {
            // secondaryBiomass contains in kg how much large herbivore mass there was available in the beginning of the year
        	// edibleMeat contains in kg how much edible meat there is
            // eatenEdibleMeat contains the kg eaten this year
            
            double eatenSecondaryBiomass = eatenEdibleMeat * (10.0/6.0);  // convert back to biomass,  60% of herbivores is edible
            double fractionEatenSecondaryBiomass = 0.0;
            
            if (secondaryBiomass > 0)
            {
                fractionEatenSecondaryBiomass = eatenSecondaryBiomass / secondaryBiomass;
            }
            
            remainingBiomassPercentage *= (1-fractionEatenSecondaryBiomass);
            remainingBiomassPercentage += recoverPercentage;  // absolute value, not relative to last year
            if (remainingBiomassPercentage >100) remainingBiomassPercentage = 100;
            
            if (Constants.ZERO_MODEL_DISABLE_ENERGY)
            {
            	secondaryBiomass = 500; // random default value
            }
            else
            {
            	if (Constants.ENERGY_CONTINUOUS)
            	{
            		secondaryBiomass = calculateSecondaryBiomassAccordingToMcNaughton();
            	}
            	else
            	{
            		secondaryBiomass = calculateSecondaryBiomassAccordingToKellyAndBinford();
            	}
            	
            }
            
            if (remainingBiomassPercentage < 100.0)
            {
            	Constants.DEBUG_PRINT_LEVEL("Secondary biomass:" + Math.round(secondaryBiomass) + " remainingBiomassPercentage: " + remainingBiomassPercentage + "  fE:" + fractionEatenSecondaryBiomass + " " + this, 15);
            }
            secondaryBiomass *= (remainingBiomassPercentage / 100.0);
            
            //double fraction = Math.min(1.0, (recoverPercentage / 100.0) + (1.0 - fractionEatenSecondaryBiomass));
            
            //if (fraction < 1.0)
            //{
//            	Constants.DEBUG_PRINT_LEVEL("Secondary biomass:" + secondaryBiomass + " fraction: " + fraction + "  fE:" + fractionEatenSecondaryBiomass + " " + this, 15);
            //}
            
            // take into account eaten amount from last year
            //secondaryBiomass *= fraction;    //  <= 1
        	
        	// of the remaining kilograms, only 50% is available as large herbivores
        	secondaryBiomass /= 2.0;
        	
        	// for each kg herbivore, only 60% is edible (rest is bone, etc)
        	edibleMeat = 0.6 * secondaryBiomass;
        	
        	
        	// hominins eat secondary biomass, consumption of primary is ignored for now
        	// this secondary biomass recovers slowly, with a maximum of the above calculated value
        	// so it is not PP that is eaten, but SB
        	// carrying capacity is a wrong word, since that refers to PP
        	
        	// for the new year, reset eaten meat
        	eatenEdibleMeat = 0.0;

        }
        
        ShowValue();
        
    }	// updateHabitatCell
  
    public void ShowValue()
    {
        // visibleValueLayer displays height, climate type, available energy, number of deads or (in the last timestep) the number of visits
		int visibleType = HSUtils.getVisualValueStyle();
		
		if (visibleType == Constants.VISIBLE_LAYER_TOPOGRAPHY)
		{
        	visibleValueLayer.set(height, x, y);
		} else if (visibleType == Constants.VISIBLE_LAYER_CLIMATE)
		{
			visibleValueLayer.set(getClimateType(), x, y); 
		} else if (visibleType == Constants.VISIBLE_LAYER_ENERGY)
		{
			visibleValueLayer.set(secondaryBiomass, x, y); 
		} else if (visibleType == Constants.VISIBLE_LAYER_DEAD)
		{
			visibleValueLayer.set(numberOfDeads, x, y); 
		} else
		{
			visibleValueLayer.set(getNumberOfVisits(), x, y);
		}
		
    }	// ShowValue
    
    public double calculateSecondaryBiomassAccordingToKellyAndBinford() 
    {
    	double result = 0.0;
        final double precipitation = getPrecipitation();
        final double MWM = getTemperatureWarmestMonth();
        final double MCM = getTemperatureColdestMonth();
        double effectiveTemp = ((18 * MWM) - (10*MCM)) / (MWM-MCM + 8);
        double a,b;

    	// determine (potentially new) climate type
    	// use precipitation to decide between grassland and forest in cool and warm climate types
    	climateType = Constants.CLIMATE_TYPE_EQUATORIAL;  // default
		a = 0.0489; b = 0.0004222;
    	
    	if (effectiveTemp < 10.0) 
    	{
    		this.climateType = Constants.CLIMATE_TYPE_POLAR;
    		a = 0.2333; b = 0.0006667;
    	}
    	else if (effectiveTemp < 12.5) 
    	{
       		if (precipitation <= precipitationColdAridLevel)
    		{
        		this.climateType = Constants.CLIMATE_TYPE_POLAR;
        		a = 0.2333; b = 0.0006667;
    		}
       		else
       		{
       			climateType = Constants.CLIMATE_TYPE_BOREAL;
       			a = 0.0400; b = 0.0002500;
       		}
    	}
    	else if (effectiveTemp < 14.56) 
    	{
       		if (precipitation <= precipitationColdAridLevel)
    		{
        		climateType = Constants.CLIMATE_TYPE_COOL_GRASS;
    			//a = 0.3800; b = 0.0043750;
    			a = 0.2250; b = 0.0037500;  // the tropical savanna parameters
    		}
    		else
    		{
        		climateType = Constants.CLIMATE_TYPE_COOL_FOREST;
    			a = 0.0371; b = 0.0002857;
    		}
    	}
    	else if (effectiveTemp < 16.62)
    	{
       		if (precipitation <= precipitationWarmAridLevel)
    			
    		{
        		climateType = Constants.CLIMATE_TYPE_WARM_GRASS;
    			a = 0.1200; b = 0.0008333;
    		}
    		else
    		{
        		climateType = Constants.CLIMATE_TYPE_WARM_FOREST;
    			a = 0.0400; b = 0.0005333;
    		}
    	}
    	else if (effectiveTemp < 18.16)
    	{
    		climateType = Constants.CLIMATE_TYPE_SUBTROPICAL;
    		a = 0.2250; b = 0.0037500;
    	}
    	else if (effectiveTemp < 22.58) 
    	{
    		climateType = Constants.CLIMATE_TYPE_TROPICAL;
    		a = 0.0457; b = 0.0003429;
    	} // else this is equatorial, the default
    	

    	// the order in which the factors are applied can make huge differences
    	double PP = calculatePrimaryProductivity();
    	
    	// PP is per cell (10000m^2) and in grams
    	result = ((PP * b)/a)/1000;
    	
    	return result;
    	
    }	// calculateSecondaryBiomassAccordingToKellyAndBinford
    

    public double calculateSecondaryBiomassAccordingToMcNaughton() 
    {
    	double result = 0.0;
        final double precipitation = getPrecipitation();
        final double MWM = getTemperatureWarmestMonth();
        final double MCM = getTemperatureColdestMonth();
        double effectiveTemp = ((18 * MWM) - (10*MCM)) / (MWM-MCM + 8);
        //double a,b;

    	// determine (potentially new) climate type
    	// use precipitation to decide between grassland and forest in cool and warm climate types
    	//climateType = Constants.CLIMATE_TYPE_EQUATORIAL;  // default
		//a = 0.0489; b = 0.0004222;
    	
    	if (effectiveTemp < 10.0) 
    	{
    		this.climateType = Constants.CLIMATE_TYPE_POLAR;
    		//a = 0.2333; b = 0.0006667;
    		//Constants.DEBUG_PRINT_LEVEL("Polar climate " + this + " precip: " + precipitation, 10 );
    	}
    	else if (effectiveTemp < 12.5) 
    	{
       		if (precipitation <= precipitationColdAridLevel)
    		{
        		this.climateType = Constants.CLIMATE_TYPE_POLAR;
        		//a = 0.2333; b = 0.0006667;
        		//Constants.DEBUG_PRINT_LEVEL("Polar climate " + this + " precip: " + precipitation, 10 );
    		}
       		else
       		{
       			climateType = Constants.CLIMATE_TYPE_BOREAL;
       			//a = 0.0400; b = 0.0002500;
       			//Constants.DEBUG_PRINT_LEVEL("Boreal climate " + this + " precip: " + precipitation, 10 );
       		}
    	}
    	else if (effectiveTemp < 14.56) 
    	{
//    		if (precipitation <= Constants.CLIMATE_PRECIPITATION_MODIFIER_COLD_ARID)
       		if (precipitation <= precipitationColdAridLevel)
    		{
        		climateType = Constants.CLIMATE_TYPE_COOL_GRASS;
    			//a = 0.3800; b = 0.0043750;
    			//a = 0.2250; b = 0.0037500;  // the tropical savanna parameters
    		}
    		else
    		{
        		climateType = Constants.CLIMATE_TYPE_COOL_FOREST;
    			//a = 0.0371; b = 0.0002857;
    		}
    	}
    	else if (effectiveTemp < 16.62)
    	{
//    		if (precipitation <= Constants.CLIMATE_PRECIPITATION_MODIFIER_WARM_ARID)
       		if (precipitation <= precipitationWarmAridLevel)
    			
    		{
        		climateType = Constants.CLIMATE_TYPE_WARM_GRASS;
    			//a = 0.1200; b = 0.0008333;
    		}
    		else
    		{
        		climateType = Constants.CLIMATE_TYPE_WARM_FOREST;
    			//a = 0.0400; b = 0.0005333;
    		}
    	}
    	else if (effectiveTemp < 18.16)
    	{
    		climateType = Constants.CLIMATE_TYPE_SUBTROPICAL;
    		//a = 0.2250; b = 0.0037500;
    	}
    	else if (effectiveTemp < 22.58) 
    	{
    		climateType = Constants.CLIMATE_TYPE_TROPICAL;
    		//a = 0.0457; b = 0.0003429;
    	} // else this is equatorial, the default
    	

    	// the order in which the factors are applied can make huge differences
    	double PP = calculatePrimaryProductivity();
    	
    	// alternative according to McNaughton
    	double PPenergy = ((PP/10000) / 1000) * 3000;
    	double secondaryBiomassMcNaughton = Math.pow(10.0, (1.52*Math.log10(PPenergy)-4.79));
    	secondaryBiomassMcNaughton = Math.max(0, (secondaryBiomassMcNaughton * 10000)/3000);    // >= 0
    	
    	result = secondaryBiomassMcNaughton;
    			
    	return result;
    	
    }   // calculateSecondaryBiomassAccordingToMcNaughton
    

    /*
     * Primary productivity according to Lieth (1973)
     *    Values minimum of PP based on temp and PP based on precip
     *    Values in gram per gridcell
     */
    private double calculatePrimaryProductivity() 
    {
    	double pp = 0.0;
    	double ppT = 0.0;
    	double ppP = 0.0;
    	final double T = getTemperature();
    	final double P = getPrecipitation();
    	
    	// calculate Primary Productivity based on Temp
    	ppT = (3000 / (1 + Math.exp(1.315-0.119*T)));

    	// calculate Primary Productivity based on Precipitation
    	ppP = 3000 * (1 - Math.exp(-0.000664*P));

    	// use minimum
    	pp = Math.min(ppT,  ppP);
    	pp = Math.max(0, pp);   // >=0
    	
    	// for the whole grid cell: gridsizeX * gridsizeY
    	// pp = pp * ((10000 * 10000) / 1000 );
    	pp = pp * 100000;
    	
    	return pp;
    	
    }	// calculatePrimaryProductivity

    /*
     * capacity consumed this time step
     * 		multiple groups can consume
     */
    public void consumeCalories(final double eatenCalories) 
    {
    	eatenEdibleMeat += eatenCalories / HomininSpaceContextBuilder.getCaloriesPerKgMeat();   //  in kg
    	addVisit();
    	
    }	// consumeCalories
    
    public double getEnvironmentCalories()
    {
		// find total capacity for the environment of h (exclude this capacity!)
		double calories = 0.0;

		for (final GridCell<HabitatCell> act : theNeighborhood) 
		{
	        for (final Object object : act.items())
	        {
                if (object instanceof HabitatCell) 
                {
                    HabitatCell nh = (HabitatCell) object;
                   	// potential target area
                    double groupsToShareWith = 1 + nh.getForagingGroupCount();
                    calories += nh.getAvailableCalories() / groupsToShareWith;
                }
	        }
		}
		
		return calories;
		
    }	// getEnvironmentCalories
        
    // this is the number of calories in edible large ungulates kilogrammes
    public double getAvailableCalories() 
    {
    	return Math.max(0, (edibleMeat - eatenEdibleMeat) * HomininSpaceContextBuilder.getCaloriesPerKgMeat());
    }	// getAvailableCalories

    public int getHabitatType()
    {
    	return habitatType;
    }	// getHabitatType
    
    public void setHabitatType( int aHabitatType )
    {
    	habitatType = aHabitatType;
    }	// setHabitatType
    
    public double getHeigth() 
    {
        return height;
    }	// getHeigth
    
    public int getXPos() 
    {
        return x;
    }	// getXPos
    
    public int getYPos() 
    {
        return y;
    }	// getYPos
    
    public int getClimateType() 
    {
        return climateType;
    }	// getClimateType
    
    public boolean getIsHomeRange() 
    {
		return isHomeRange;
	}	// getIsHomeRange
    
	public void setIsHomeRange(boolean isHomeRange) 
	{
		this.isHomeRange = isHomeRange;
	}	// setIsHomeRange

    /*
     * setHeight
     *   newDefaultHeight = true when a fake height, for instance a depth, is set
     */
    private void setHeigth( double aHeight, boolean newDefaultHeight ) {
    	if (newDefaultHeight)
    	{
    		if (cellData != null)
    		{
    			cellData.setHeight(aHeight);
    		}
    	}
    	
    	height = aHeight;
    	
        if (height < Constants.SEA_LEVEL_HEIGHT) 
        {
        	// we are (now, maybe due to changing climate) below sea level, set new type if applicable
   			if (this.cellData.getType() != Constants.HABITAT_TYPE_LAKE)
   			{
                habitatType = Constants.HABITAT_TYPE_SEA;
   			}
        }
        else
        {
            if (height < Constants.PLAIN_HEIGHT) {
                habitatType = Constants.HABITAT_TYPE_PLAIN;
            }
            else
                if (height < Constants.HILL_HEIGHT) {
                    habitatType = Constants.HABITAT_TYPE_HILL;
                }
                else
                	habitatType = Constants.HABITAT_TYPE_MOUNTAIN;
        }
        
    }	// setHeigth
    
    public double getTemperatureColdestMonth() {
        return cellData.getTemperatureColdestMonthInterpolated();
    }
    public double getTemperatureWarmestMonth() {
        return cellData.getTemperatureWarmestMonthInterpolated();
    }
    public double getTemperature() {
        return cellData.getTemperatureInterpolated();
    }
    public double getPrecipitation() {
        return cellData.getPrecipitationInterpolated();
    }
    public double getCurrentPrecipitation() {
        return cellData.getCurrentPrecipitation();
    }
    public double getCurrentTemperature() {
        return cellData.getCurrentTemperature();
    }
    public double getCurrentWarmestMonthTemperature() {
        return cellData.getCurrentWarmestMonthTemperature();
    }
    public double getCurrentColdestMonthTemperature() {
        return cellData.getCurrentColdestMonthTemperature();
    }

    public void setCurrentPrecipitation(double currentPrecipitation) {
        cellData.setCurrentPrecipitation(currentPrecipitation);
    }
    public void setCurrentTemperature(double currentTemperature) {
        cellData.setCurrentTemperature(currentTemperature);
    }
    public void setCurrentTemperatureWarmestMonth(double currentTemperature) {
        cellData.setCurrentWarmestMonthTemperature(currentTemperature);
    }
    public void setCurrentTemperatureColdestMonth(double currentTemperature) {
        cellData.setCurrentColdestMonthTemperature(currentTemperature);
    }

    // for density plot
    public void addVisit()
    {
    	if (this.habitatType != Constants.HABITAT_TYPE_SEA)
    	{
    		numberOfVisits++;
    	}
    }
    public int getNumberOfVisits()
    {
    	return numberOfVisits;
    }
    public void addDeads( int aNumberOfDeads )
    {
    	numberOfDeads += aNumberOfDeads;
    }
    public void addCheckpoint( Checkpoint c)
    {
    	if (theCheckpoints == null)
    	{
    		theCheckpoints = new ArrayList<Checkpoint>();
    	}
   		theCheckpoints.add(c);
    }
    public void setPresence( int id, int type)
    {
    	if (theCheckpoints != null)
    	{
    		for (int i=0; i<theCheckpoints.size(); i++)
    		{
    			theCheckpoints.get(i).setPresence(id, type);
    			if (isHomeRange)
    			{
    				theCheckpoints.get(i).setPresenceAsHomeRange();
    			}
    		}
    	}
    }
    public void increaseForagingGroupCount()
    {
    	if ((this.habitatType != Constants.HABITAT_TYPE_SEA) &&
    		(this.habitatType != Constants.HABITAT_TYPE_LAKE))
    	{
        	foragingGroups++;
    	}
    }
    public void decreaseForagingGroupCount()
    {
    	if (foragingGroups > 0)
    		foragingGroups--;
    }
    public int getForagingGroupCount() {
        return foragingGroups;
    }
    public int getForagingGroupLastYearCount() {
        return foragingGroupsLastYear;
    }
    public int getForagingGroupLastLastYearCount() {
        return foragingGroupsLastLastYear;
    }

    public void setEnvironmentList(List<GridCell<HabitatCell>> aList)
    {
    	theNeighborhood = aList;
    }
    
    @Override
    public String toString() {
        return String.format(
           "HabitatCell (x=%d, y=%d), Ti:%.2f, GT:%.2f Pi:%.2f, Climate: %d, SB=%.2f, height=%.2f", x, y, cellData.getTemperatureInterpolated(),
           HomininSpaceContextBuilder.getGlobalTemperatureOffset(), this.cellData.getPrecipitationInterpolated(), climateType, secondaryBiomass, height);
    }

}
