/*
 Copyright 2011-2018 Fulco Scherjon

This file is part of the HomininSpace modelling and simulation system.

HomininSpace is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, any version 3.
HomininSpace is distributed WITHOUT ANY WARRANTY under the GNU General Public License.
HomininSpace comes with a copy of this GNU General Public License.  If not, see <http://www.gnu.org/licenses/>.

*/


package HomininSpace.agents;


import java.util.ArrayList;
import java.util.List;

import repast.simphony.query.space.grid.GridCell;
import repast.simphony.query.space.grid.GridCellNgh;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.grid.Grid;
import repast.simphony.space.grid.GridPoint;
import repast.simphony.util.SimUtilities;
import HomininSpace.HomininSpaceContextBuilder;
import HomininSpace.common.Constants;
import HomininSpace.common.HSUtils;



public class HomininFactory {

	private int 					ID;					// unique identifier, the type of hominin to produce
	private String 					name;     			// readable identification
	private String 					type;     			// type in characters (to be used in shape files)
	private int						posX = -1;
	private int						posY = -1;
	private int 					createdGroups = 0; 
	private int						id = -1;
	private Grid<Object> 			grid = null;	
	public boolean					groupsCanSettle = Constants.STATIC_DISPERSAL;
	public double					caloriesRequired = 0.0;
	
	// constructor
	public HomininFactory( int id )
	{
		this.id = id;
		name = "F" + id;
		type = "1";   // default (note: strings because it appears that only strings are passed from tab to shape files)
	}

	public void initialize( Grid<Object> grid )
	{
		this.grid = grid;
		
		if (null == grid) {
			throw new IllegalStateException("Factory: Cannot locate grid in context.");
		}
	}
	
	/*
	 * Factory production, called each turn (max 1 group per turn)
	 * 		when parameter forced=true production occurs regardless of environmental conditions
	 */
	public void produce(boolean forced) 
	{
		boolean roomForGroup 			= true;
		int scanningRange 				= Constants.FACTORY_SCANNING_RANGE;
		//int localGroupCount 			= 0;		// the number of groups in the scanning range of the factory
		int localHabitableGridcellCount	= 0;		// the number of grid cells that can accommodate new groups
		//double caloriesPresent			= 0.0;		// amount of calories in the foraging area around the factory
        final HomininGroup newGroup; 				//new HomininGroup(HomininSpaceContextBuilder.getNewAgentId());
		final GridPoint 	currentLocation 		= grid.getLocation(this);
		
		if (!forced)
		{
			// check environment if group can be created
	        final List<GridCell<HomininGroup>> groupNeighborhood = new GridCellNgh<HomininGroup>(
	                                grid, currentLocation, HomininGroup.class,
	                                scanningRange, scanningRange).getNeighborhood(true);
	        
	        // if any hungry group is in the area, do not produce 
	        searchHungryGroup:
			for (final GridCell<HomininGroup> act : groupNeighborhood) 
			{
				if ((act.size() > 0) && roomForGroup) 
				{   
					// size is number of agents at this gridcell
					for (final Object object : act.items()) 
					{
		                if (object instanceof HomininGroup) 
		                {
		                	//localGroupCount++;
		                	HomininGroup h = (HomininGroup) object;
		                	if (h.isHungry())
		                	{
		                		roomForGroup = false;
			                	break searchHungryGroup;
		                	}
		                }
			        }
				}
			}
			
			if (roomForGroup)
			{
/*				
				// check if there are grid cells that can accommodate groups
				final List<GridCell<HabitatCell>> theNeighborhoodToMoveTo = new GridCellNgh<HabitatCell>(grid,
						currentLocation, HabitatCell.class, scanningRange, scanningRange).getNeighborhood(true);
				for (final GridCell<HabitatCell> act : theNeighborhoodToMoveTo) 
				{
			        for (final Object object : act.items())
			        {
		                if (object instanceof HabitatCell) 
		                {
		                	HabitatCell h = (HabitatCell) object;
		                    if ((h.getHabitatType() != Constants.HABITAT_TYPE_SEA) && 
		                    	(h.getHabitatType() != Constants.HABITAT_TYPE_LAKE) &&
		                    	(!h.getIsHomeRange()) &&
		                    	(h.getForagingGroupCount() == 0 ))
		                    {
		                    	// count the number of available grid cells
		        				localHabitableGridcellCount++;
				                caloriesPresent += h.getAvailableCalories();
		                    }
		                }
			        }
				}

				if (localHabitableGridcellCount == 0)
				{
					roomForGroup = false;
				}
*/				
/*				
				// check for group density
				double density = 100;
				if (localHabitableGridcellCount > 0)
				{
					density = (localGroupCount / localHabitableGridcellCount) * 100;
					if (density > 5)
					{
						Constants.DEBUG_PRINT_LEVEL("Group density too high around " + this.name + "  (" + density + ")", 5);
						roomForGroup = false;
					}
				}
				else
				{
					roomForGroup = false;
				}
*/				
			}
		}
		
		if (roomForGroup)
		{
			//final GridCell chosenFreeCell = null;
			GridPoint newGridPoint = null;
			
			// check if there are grid cells that can accommodate groups
			final List<GridCell<HabitatCell>> theNeighborhoodToMoveTo = new GridCellNgh<HabitatCell>(grid,
					currentLocation, HabitatCell.class, scanningRange, scanningRange).getNeighborhood(true);
	        SimUtilities.shuffle(theNeighborhoodToMoveTo, RandomHelper.getUniform());

	        searchNewLocation:
			for (final GridCell<HabitatCell> act : theNeighborhoodToMoveTo) 
			{
		        for (final Object object : act.items())
		        {
	                if (object instanceof HabitatCell) 
	                {
	                	HabitatCell h = (HabitatCell) object;
	                    if ((h.getHabitatType() != Constants.HABITAT_TYPE_SEA) && 
	                    	(h.getHabitatType() != Constants.HABITAT_TYPE_LAKE) &&
	                    	(!h.getIsHomeRange()) &&
	                    	(h.getForagingGroupCount() == 0 ))
	                    {
	                    	// compute the available kcalories in the area
	                    	double kCal = 0.0; 
	            			final List<GridCell<HabitatCell>> theNeighborhoodResources = new GridCellNgh<HabitatCell>(grid,
	            					act.getPoint(), HabitatCell.class, HomininSpaceContextBuilder.getMaxForagingRange(), HomininSpaceContextBuilder.getMaxForagingRange()).getNeighborhood(true);

	            			for (final GridCell<HabitatCell> actRes : theNeighborhoodResources) 
	            			{
	            		        for (final Object objectRes : actRes.items())
	            		        {
	            	                if (objectRes instanceof HabitatCell) 
	            	                {
	            	                	HabitatCell hRes = (HabitatCell) object;
	            	                	kCal += hRes.getAvailableCalories();
	            	                }
	            		        }
	            			}
	            	        if (kCal > caloriesRequired)        	
	            	        {
	            	        	// the first cell that can accommodate a new group in the scanning range
	            	        	newGridPoint = act.getPoint();
	            	        	break searchNewLocation;
	            	        }
	                    }
	                }
		        }
			}

			if (newGridPoint == null)
			{
				roomForGroup = false;
			}
			else
			{
				/*
				 * Create new human group
				 */
				newGroup = HomininSpaceContextBuilder.getNewHomininGroup(); //new HomininGroup(HomininSpaceContextBuilder.getNewAgentId());
				newGroup.setHomininType(this.id);
				
				// from now on we know how many kcal are required for a new group, so that they do not die directly after producing
				caloriesRequired = newGroup.getCaloriesRequired();
	
		        // Add the new group to the context and to the grid
		        // We have our new GridPoint to move to, so locate new group
		        int newX = newGridPoint.getX(); 
		        int newY = newGridPoint.getY();
		        if (posX < 30)
		        {
		        	// Iberia: send to the right, so set old pos to the left
		        	newGroup.setPosXY(newX, newY, posX+10, posY+1);  
		        }
		        else
		        {
		        	// Italy: send to the left, so set old pos to the right
		        	newGroup.setPosXY(newX, newY, posX+6, posY+1);  
		        }
		        // the first turn they will not consume 
		        newGroup.setForagingRange();
		        newGroup.setCanSettle(groupsCanSettle);
		        newGroup.setAge(HomininSpaceContextBuilder.getYearsBeforeGroupMaturity());  // these groups can immediately settle or join
		       
				Constants.DEBUG_PRINT_LEVEL( "Factory " + this.name + " produces group " + newGroup.getId() +  "  x:" + newX + "  y:"+newY, 10);
				
				createdGroups++;
			}
		}
		
		if (!roomForGroup)
		{
			Constants.DEBUG_PRINT_LEVEL("Factory " + this + " does NOT produce this turn", 15);
		}
	}  // step
	
	/**
	 * @return the iD
	 */
	public int getID() 
	{
		return ID;
	}

	/**
	 * @param id; the ID to set
	 */
	public void setID(int id) 
	{
		this.ID = id;
	}

	/**
	 * @return the name
	 */
	public String getName() 
	{
		return name;
	}

	/**
	 * @param name; the name to set
	 */
	public void setName(String name) 
	{
		this.name = name;
	}

	/**
	 * @return the name
	 */
	public String getType() 
	{
		return this.type;
	}

	/**
	 * @param name; the name to set
	 */
	public void setType(String aType) 
	{
		this.type = aType;
	}
	
	public void setCanSettle(boolean aValue) 
	{
		this.groupsCanSettle = aValue;
	}
	
	public int getCreatedGroups()
	{
		return createdGroups;
	}
	
    public int getPosX() {
		final GridPoint location = grid.getLocation(this);
		posX = location.getX();
        return posX;
    }
    public int getPosY() {
		final GridPoint location = grid.getLocation(this);
		posY = location.getY();
        return posY;
    }
	public void setPosXY(final int x, final int y) {
		grid.moveTo(this, x, y);
		this.posX = x;
		this.posY = y;
	}

    @Override
    public String toString() 
    {
            return String.format("%s [x=%d, y=%d]", this.name, posX, posY);
    }

}
