/*
 Copyright 2011-2018 Fulco Scherjon

This file is part of the HomininSpace modelling and simulation system.

HomininSpace is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, any version 3.
HomininSpace is distributed WITHOUT ANY WARRANTY under the GNU General Public License.
HomininSpace comes with a copy of this GNU General Public License.  If not, see <http://www.gnu.org/licenses/>.

*/

// !!!!!!   look at the TODO for zero model reverse 


package HomininSpace.agents;

/*
 * Class HomininGroup - implements Hominingroups in the HomininSpace simulation system
 * 
 * @author: F. Scherjon
 * @version $Id: HomininGroup.java
 * 
 */

import HomininSpace.HomininSpaceContextBuilder;
import HomininSpace.common.*;

import java.util.List;

import repast.simphony.query.space.grid.GridCell;
import repast.simphony.query.space.grid.GridCellNgh;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.grid.Grid;
import repast.simphony.space.grid.GridPoint;
import repast.simphony.util.ContextUtils;
import repast.simphony.util.SimUtilities;


public class HomininGroup {
	
	private HabitatCell currentHabitatCell 		= null;
	private GridPoint 	currentLocation 		= null;
	private int 		homininType				= Constants.CULTURAL_TYPE_0;
	private Grid<Object> grid 					= null;		

	private double 		sizePreFertile 			= 0;
	private double 		sizeFertile 			= 0;
	private double 		sizePostFertile 		= 0;
	private int 		sizePreFertileCategory	= 0;
	private int 		sizeFertileCategory		= 0;
	private double 		consumptionPreFertile 	= 0;
	private double 		consumptionFertile 		= 0;
	private double 		consumptionPostFertile 	= 0;
	
	private double 		birthsPerFemale 		= 0;
	private double 		deathsPreFertile 		= 0;
	private double 		deathsFertile 			= 0;
	private double 		deathsPostFertile 		= 0;
	
	private int 		foragingRange 			= 1;		// in grid cells
	private double		minimumTemperature		= 0.0;		// in degrees Celcius
	
	private int 		posX 					= 1;
	private int 		posY 					= 1;
	private int 		id 						= -1;

	private boolean 	isGroupAlive			= true;
	private boolean 	isGroupHungry 			= false;
	private int			yearsHungry				= 0;
	private int			dieCountdown			= 0;		// number of years a dead group is visible (Interface)
	private boolean		isGroupSmall			= false;
	private boolean		isGroupSettled			= false;
	private int			age						= 1;
	private boolean		hungryYears				= false;	// indicates if previous year was also hungry
	private boolean		canSettle				= false;
	
	/*
	 * Creates a new instance of <code>HomininGroup</code> 
	 */
	public HomininGroup( Grid<Object> aGrid ) 
	{
		super();
		
		grid 				= aGrid;
		
		if (null == grid) 
		{
			throw new IllegalStateException("HomininGroup: Cannot locate grid in context.");
		}
	}

	/*
	 * Initialize hominin group
	 */
	public void initialize( int anId )
	{
		id = anId;
		
		//careful: location is not defined yet!!!
		currentHabitatCell 		= null;
		currentLocation 		= null;
		homininType				= Constants.CULTURAL_TYPE_0;

		sizePreFertile 			= 0; //Constants.INITIAL_SIZE_PREFERTILE;
		sizeFertile 			= 0; //Constants.INITIAL_SIZE_FERTILE;
		sizePostFertile 		= 0; //Constants.INITIAL_SIZE_POSTFERTILE;
		
		birthsPerFemale 		= HomininSpaceContextBuilder.getBirthsPercentage()/100.0; //Constants.BIRTHS_PERCENTAGE/100.0;  // per year
		sizePreFertileCategory	= HomininSpaceContextBuilder.getSizePreFertileGroup();
		sizeFertileCategory		= HomininSpaceContextBuilder.getSizeFertileGroup();
		deathsPreFertile 		= HomininSpaceContextBuilder.getDeathsPreFertilePercentage()/100.0;
		deathsFertile 			= HomininSpaceContextBuilder.getDeathsFertilePercentage()/100.0;
		deathsPostFertile 		= HomininSpaceContextBuilder.getDeathsPostFertilePercentage()/100.0;
		consumptionPreFertile 	= HomininSpaceContextBuilder.getConsumptionPreFertile();
		consumptionFertile 		= HomininSpaceContextBuilder.getConsumptionFertile();
		consumptionPostFertile 	= HomininSpaceContextBuilder.getConsumptionPostFertile();
		
		posX 					= 1;
		posY 					= 1;
		age						= 1;

		setAlive(true);
		isGroupHungry			= false;
		isGroupSettled			= false;
		yearsHungry				= 0;

		minimumTemperature 		= HomininSpaceContextBuilder.getMinimumTemperature();  // minimum temperature needed to survive
		foragingRange			= 1;   // default value to start with
		
		canSettle				= HomininSpaceContextBuilder.getStaticDispersal();
	}  // initialize

	/*
	 * Implementation of group movement
	 * 		Settle if possible otherwise look for the best destination grid cell (patch) within 2*scanningrange
	 */
	public void move() 
	{
		age++; // count the years this group is alive

		// if not settled yet, first check if it is possible settle, otherwise move on
		if (!isGroupSettled)
		{
			// first, check if this group can settle at all
			boolean settlingPosible = canSettle;
			
			// do not settle if there is not enough food around
			if (isGroupHungry)
			{
				settlingPosible = false;
			}
			if (age < HomininSpaceContextBuilder.getYearsBeforeGroupMaturity())
			{
				settlingPosible = false;
			}

			if (settlingPosible)
			{
				// Look for humans that are near; if there are one or more settled then settling is not possible
				final List<GridCell<HomininGroup>> theNeighborhood = new GridCellNgh<HomininGroup>(grid,
						currentLocation, HomininGroup.class, foragingRange, foragingRange).getNeighborhood(false);
		
				// find all HomininGroups within scanning range
				search:
				for (final GridCell<HomininGroup> act : theNeighborhood) 
				{
			        for (final Object object : act.items()) 
			        {
		                if (object instanceof HomininGroup) 
		                {
		                	HomininGroup h = (HomininGroup) object;
		                	if (h.isSettled())
		                	{
		                		settlingPosible = false;
		                		break search;   // break the outer for loop; optimization
		                	}
		                }
			        }
				}
				
				// do not settle near the border, because of insufficien resources
				int x = currentLocation.getX();
				int y = currentLocation.getY();
				if ((x <= 2) || (x >= HomininSpaceContextBuilder.getGridSizeX()-3) || (y <= 2) || (y >= HomininSpaceContextBuilder.getGridSizeY()-3))
				{
	            	settlingPosible = false;
				}
			}
			
			if (settlingPosible)
			{
				// settle this group here
				isGroupSettled = true;
			}
			else
			{
				// move this group to a new position (include center which might be the best position (but probably not since most consumption was from the center))
				// for this calculation, use the foraging range of previous year
				
				final List<GridCell<HabitatCell>> theNeighborhoodToMoveTo = new GridCellNgh<HabitatCell>(grid,
						currentLocation, HabitatCell.class, 2*(foragingRange), 2*(foragingRange)).getNeighborhood(true);  // was foragingRange,	foragingRange // was 4 4 // was 6 6

				// create random order to avoid movement bias from the neighborhood function
				SimUtilities.shuffle(theNeighborhoodToMoveTo, RandomHelper.getUniform());

				// find maximum capacity within foraging range as the most attractive destination
				double bestCapacity = -10000000000.0;
				HabitatCell bestHabitatCell = null;
				int bestCapacityGroupCount = 100000;  // number of groups that foraged at the best destination last year
				int groupsToShareWith = 0;  // number of groups that are going to share in a grid cell
				
				if (Constants.ZERO_MODEL_RANDOM_WALK)   // random move, get first available grid cell
				{
					searchNeighbourhood:
					for (final GridCell<HabitatCell> act : theNeighborhoodToMoveTo) 
					{
				        for (final Object object : act.items())
				        {
			                if (object instanceof HabitatCell) 
			                {
			                	HabitatCell h = (HabitatCell) object;
			                    if (	((h.getHabitatType() != Constants.HABITAT_TYPE_SEA) || Constants.ZERO_MODEL_DISABLE_WATER) && 
			                    		((h.getHabitatType() != Constants.HABITAT_TYPE_LAKE) || Constants.ZERO_MODEL_DISABLE_WATER) &&
			                    		(!h.getIsHomeRange() || (h == currentHabitatCell))) // occupied by another group or my own cell
			                    {
			                    	bestHabitatCell = h;
			                    	break searchNeighbourhood;
			                    }
			                }
				        }
					}
					
				}
				else
				{
					for (final GridCell<HabitatCell> act : theNeighborhoodToMoveTo) 
					{
				        for (final Object object : act.items())
				        {
			                if (object instanceof HabitatCell) 
			                {
			                    HabitatCell h = (HabitatCell) object;
			                    // new location should not be water, not near the (closed) borders and not already occupied
			                    // near the borders groups die rapidly, although the location looks good
			                    if (	((h.getHabitatType() != Constants.HABITAT_TYPE_SEA) || Constants.ZERO_MODEL_DISABLE_WATER) && 
			                    		((h.getHabitatType() != Constants.HABITAT_TYPE_LAKE) || Constants.ZERO_MODEL_DISABLE_WATER) &&
//			                    if (	((h.getHabitatType() != Constants.HABITAT_TYPE_SEA)) && 
//					                    ((h.getHabitatType() != Constants.HABITAT_TYPE_LAKE)) &&
			    						//(((HomininSpaceContextBuilder.getOpenBorders() == true) && (h.getXPos() > 2) && (h.getXPos() < HomininSpaceContextBuilder.getGridSizeX()-3) && (h.getYPos() > 2) && (h.getYPos() < HomininSpaceContextBuilder.getGridSizeY()-3)) || (HomininSpaceContextBuilder.getOpenBorders() == false)) &&
			                    		(!h.getIsHomeRange() || (h == currentHabitatCell))) // occupied by another group or my own cell
			                    {
			                    	// potential target area to move to
			                    	// to avoid death traps we compare cells and their immediate environments with each other, not just cells
				                    double cap = h.getAvailableCalories() + h.getEnvironmentCalories();
				                    
				                    groupsToShareWith = 1 + h.getForagingGroupCount();
				                    
				                    if (groupsToShareWith < bestCapacityGroupCount)
				                    {
			                    		bestHabitatCell = h;
			                    		bestCapacity = cap;
			                    		bestCapacityGroupCount = groupsToShareWith; 
				                    }
				                    else if (groupsToShareWith == bestCapacityGroupCount)
				                    {
				                    	//if (cap > (bestCapacity * 1.1))  // at least 10% difference
				                    	if (cap > bestCapacity)
				                    	{
				                    		bestHabitatCell = h;
				                    		bestCapacity = cap;
				                    	}
				                    }
			                    }
			                }
				        }    				
					}
				}
				
		
				// if there is no empty location in the foraging range, the HomininGroup cannot move
				if (bestHabitatCell == null) 
				{
					// should not happen!
			        Constants.DEBUG_PRINT_LEVEL("HomininGroup would die because there is nowhere to go - " + this, 5);
//					isGroupAlive = false;
					//return;
				}
				else
				{
					// move the group
					currentHabitatCell.setIsHomeRange(false);
					
					// new GridPoint to move to, so relocate agent
					int newX = bestHabitatCell.getXPos();
					int newY = bestHabitatCell.getYPos();
					getGrid().moveTo(this, newX, newY);
					currentLocation = grid.getLocation(bestHabitatCell);
					currentHabitatCell = bestHabitatCell;
					currentHabitatCell.setIsHomeRange(true);
					currentHabitatCell.addVisit();
					
					if (HomininSpaceContextBuilder.getOpenBorders() == true)
					{
						// check if we moved off the board with open borders = remove group
						if ((newX <= 2) || (newX >= HomininSpaceContextBuilder.getGridSizeX()-3) || (newY >= HomininSpaceContextBuilder.getGridSizeY()-3))
						{
							if ((newX <= 2) || (newY==0))
							{
								HSUtils.incrementHomininGroupDeathCount( this.getHomininType(), this.getSize(), Constants.CAUSE_OF_DEATH_BORDER_IBERIA );
							}
							else
							{
								HSUtils.incrementHomininGroupDeathCount( this.getHomininType(), this.getSize(), Constants.CAUSE_OF_DEATH_BORDER );
							}
								
					        Constants.DEBUG_PRINT_LEVEL("HomininGroup dies because they moved of the edge - " + this, 10);
					        setAlive(false);
						}
					}
				}
			}
		}
		
		if (isAlive())
		{
			setForagingRange();
		}
	}

	// determine foraging range for currentHabitatCell, based on needed calories and calories found in the surrounding cells
	public void setForagingRange()
	{
		double caloriesRequired = getCaloriesRequired();
		double caloriesPresent = 0.0;
		int range = 0;   // minimum value!
		List<GridCell<HabitatCell>> theNeighborhood = null;
		
		while (caloriesPresent < caloriesRequired)
		{
			caloriesPresent = 0;
			range++;
			theNeighborhood = new GridCellNgh<HabitatCell>(getGrid(),
					currentLocation, HabitatCell.class, range, range).getNeighborhood(true);
			
			// check to see how much food is available
			for (final GridCell<HabitatCell> act : theNeighborhood) 
			{
				if (caloriesPresent < caloriesRequired)
				{
			        for (final Object object : act.items()) 
			        {
			            if (object instanceof HabitatCell) 
			            {
			                HabitatCell h = (HabitatCell) object;
			                // estimation of future consumption includes other groups in the area
			                caloriesPresent += h.getAvailableCalories()/(1+h.getForagingGroupCount());
			            }
				    }
				}
				else
				{
					break;
				}
			}
		}
		
		// found the range where enough calories are present
		if (range>HomininSpaceContextBuilder.getMaxForagingRange())
		{
			range = HomininSpaceContextBuilder.getMaxForagingRange();
		}
		foragingRange = range;
		
        // Administration of presence for calculation of available food (make a reservation in all grid cells)
		for (final GridCell<HabitatCell> act : theNeighborhood) 
		{
	        for (final Object object : act.items()) 
	        {
	            if (object instanceof HabitatCell) 
	            {
	                HabitatCell h = (HabitatCell) object;
	                h.increaseForagingGroupCount();
	            }
		    }
		}
	}
	
	/*
	 * Locate the required amount of energy from the area within foraging range and consume this
	 * 		If insufficient food is available apply modifiers to birth and death rates 
	 */
    public void eat() 
	{
    	// the actual eating
    	consume();
    	
		// administrate whether the group could eat enough
        if (isGroupHungry)
        {
        	yearsHungry++;
          	
        	// was previous year also bad, apply modifiers
        	hungryYears = true;
        	if (hungryYears)
        	{
        		// reduce newborns
	        	birthsPerFemale = birthsPerFemale / 2.0;
	        	// increase death rate; limit rate to 50%
	        	deathsPreFertile *= 2; deathsPreFertile = Math.min( 0.50, deathsPreFertile);
	        	deathsFertile *= 2; deathsFertile = Math.min( 0.50, deathsFertile);
	        	deathsPostFertile *= 2; deathsPostFertile = Math.min( 0.50, deathsPostFertile);
        	}
        	
        	hungryYears = true;
        }
        else
        {
        	// reset demographic variables instantly, now there is enough to eat
        	hungryYears = false;
        	yearsHungry = 0;
        	birthsPerFemale = HomininSpaceContextBuilder.getBirthsPercentage()/100.0;        	
        	deathsPreFertile = HomininSpaceContextBuilder.getDeathsPreFertilePercentage()/100.0;
        	deathsFertile = HomininSpaceContextBuilder.getDeathsFertilePercentage()/100.0;
        	deathsPostFertile = HomininSpaceContextBuilder.getDeathsPostFertilePercentage()/100.0;
        }
    	
        // cultural limit on growth based on group size
        if (getSize() > HomininSpaceContextBuilder.getGroupSizeLarge())
        {
        	// note: this is not cumulative, as these variables are reset above if there is enough food
        	birthsPerFemale = birthsPerFemale / 2.0;
        }
	}
	
    /*
     *  administrate Consumption, each turn required calories must be eaten
     */
    private void consume() 
	{
		double caloriesRequired = getCaloriesRequired();
		double caloriesConsumed = 0.0;
		
		isGroupHungry = true;   // until proven otherwise!
		
		// now consume the food from randomized list (not all food is eaten)
		final List<GridCell<HabitatCell>> theNeighborhood = new GridCellNgh<HabitatCell>(getGrid(),
				currentLocation, HabitatCell.class, foragingRange, foragingRange).getNeighborhood(false);
		SimUtilities.shuffle(theNeighborhood, RandomHelper.getUniform());
		
		// eat all from current home cell
		caloriesConsumed = currentHabitatCell.getAvailableCalories();
		currentHabitatCell.consumeCalories(caloriesConsumed);
		currentHabitatCell.setPresence(id, this.homininType);
		
//		Maybe sort according to productivity level?

		searchCells:
		// eat as much as needed from other cells (in random order)
		for (final GridCell<HabitatCell> act : theNeighborhood) 
		{
		        for (final Object object : act.items()) 
		        {
		            if (object instanceof HabitatCell) 
		            {
		                HabitatCell h = (HabitatCell) object;
		                
		                // we cannot eat from other home cells and we eat until enough, emptying each grid cell completely. We share with other groups if they have claimed access.
		                if (caloriesConsumed < caloriesRequired)
		                {
			                if (!h.getIsHomeRange())
			                {
			                	int groupsToShareWith = Math.max(1, h.getForagingGroupCount()-1);
				                double cal = h.getAvailableCalories() / groupsToShareWith;
				                if (cal+caloriesConsumed > caloriesRequired)
				                {
				                	cal = caloriesRequired - caloriesConsumed + 1;
				                }
				                caloriesConsumed += cal;
				                h.consumeCalories(cal);
				                h.setPresence(id, this.homininType);
			                }
		                }
		                else
		                {
		                	break searchCells;
		                }
		            }
			    }
		}
		
		if (caloriesConsumed >= caloriesRequired)
		{
			//Constants.DEBUG_PRINT_LEVEL("Enough food:" + caloriesRequired + " consumed: " + caloriesConsumed + "  " + this, 5);
			isGroupHungry = false;
		}
		else
		{
			String s = String.format("Not enough food, required:%.3f;  consumed:%.3f)", caloriesRequired, caloriesConsumed);
			Constants.DEBUG_PRINT_LEVEL(s + this, 10);
			if (caloriesConsumed < (caloriesRequired / 10))   // TODO: check if this results in too many deaths?
			{
				Constants.DEBUG_PRINT_LEVEL("HomininGroup dies because there is not enough food: " + caloriesConsumed + " ("+ caloriesRequired+ ")  " + this, 10);
		        HSUtils.incrementHomininGroupDeathCount(homininType, getSize(), Constants.CAUSE_OF_DEATH_HUNGER);
		        setAlive(false);
			}
		}
	}
    
    /*
     * Grow - HomininGroup demographics
     */
    public void grow() 
    {
	    // for each age category, grow  // we already moved and consumed capacity here
	    final double becomingFertile = sizePreFertile / sizePreFertileCategory;  // size of age category
	    final double becomingPostFertile =  sizeFertile / sizeFertileCategory; // size of fertile age category
	    final double newBorns =  (sizeFertile / 2.0) * birthsPerFemale;   // only females give birth
	   
	    final double dPre = sizePreFertile * deathsPreFertile;
	    final double dFertile = sizeFertile * deathsFertile;
	    final double dPost = sizePostFertile * deathsPostFertile;
		HSUtils.incrementHomininDeathCount(getHomininType(), dPre+dFertile+dPost);
	    
  		// administration of births and deaths
	    sizePreFertile += newBorns;
	    sizePreFertile -= becomingFertile;
	    sizeFertile += becomingFertile;
	    sizeFertile -= becomingPostFertile;
	    sizePostFertile += becomingPostFertile;

 		sizePreFertile -= dPre;
   		sizeFertile -= dFertile;
   		sizePostFertile -= dPost;
   		
        isGroupSmall = (getSize() < HomininSpaceContextBuilder.getGroupSizeSmall()) || (sizeFertile < HomininSpaceContextBuilder.getGroupSizeFertileSmall());
    }   // grow
    
	/*
	 * Try to produce a new group within this territory if that is possible and required
	 */
	public void reproduce() 
	{
		// check conditions for creating a new group
        if (getSize() > HomininSpaceContextBuilder.getGroupSizeLarge())
        {
	        double newSizePreFertile = 0;
	        double newSizeFertile = 0;
	        double newSizePostFertile = 0;

	        int newPosX = 0; 
	        int newPosY = 0;
			final List<GridCell<HabitatCell>> theNeighborhoodToMoveTo = new GridCellNgh<HabitatCell>(getGrid(),
					currentLocation, HabitatCell.class, foragingRange, foragingRange).getNeighborhood(false);
			SimUtilities.shuffle(theNeighborhoodToMoveTo, RandomHelper.getUniform());
	
			searchLocation:
			for (final GridCell<HabitatCell> act : theNeighborhoodToMoveTo) 
			{
			        for (final Object object : act.items())
			        {
		                if (object instanceof HabitatCell) 
		                {
		                    HabitatCell h = (HabitatCell) object;

		                    if (Constants.ZERO_MODEL_DISABLE_WATER == true)
		                    {
		                    	// water is no limiting factor
		                    	if (!h.getIsHomeRange())
		                    	{
		                    		// found a new location
		                    		newPosX = h.getXPos();
		                    		newPosY = h.getYPos();
		                    		break searchLocation;
		                    	}
		                    }
		                    else
		                    {
		                    	if ((h.getHabitatType() != Constants.HABITAT_TYPE_SEA) &&
		                    		(h.getHabitatType() != Constants.HABITAT_TYPE_LAKE) &&
		                    		(!h.getIsHomeRange()))
		                    	{
		                    		// found a new location
		                    		newPosX = h.getXPos();
		                    		newPosY = h.getYPos();
		                    		break searchLocation;
		                    	}
		                    }
		    	        }
			        }
			}
			
	        final HomininGroup newGroup = HomininSpaceContextBuilder.getNewHomininGroup(); // new HomininGroup(HomininSpaceContextBuilder.getNewAgentId());
	        
	        // new group size is half of the current group
	        newSizePreFertile = sizePreFertile/2;
	        newSizeFertile = sizeFertile/2;
	        newSizePostFertile = sizePostFertile/2;
	
	        sizePreFertile -= newSizePreFertile;
	        sizeFertile -= newSizeFertile;
	        sizePostFertile -= newSizePostFertile;
	        
	        // There is a new GridPoint to move to, so set new size and locate new group
	        newGroup.setPosXY(newPosX, newPosY, posX, posY);
	        newGroup.setSize(newSizePreFertile, newSizeFertile, newSizePostFertile);
	        // similar type for offspring
	        newGroup.setHomininType(getHomininType());
	        newGroup.setForagingRange();
	        HSUtils.incrementHomininOffspring();
	        
	        Constants.DEBUG_PRINT_LEVEL("Group added (from: " + posX + "," + posY + ")  id=" + newGroup+ "  at x:"+newPosX+  "  y:"+newPosY + " frange: " + foragingRange + "   by ->"+this, 10);
        }
  	}
	
	/*
	 * Merge - Check if in the foraging range there is a group to merge with, if so send merge request
	 */
	public boolean merge() 
	{
        final Grid<Object> grid = getGrid();
        boolean hasMerged = false;

        final List<GridCell<HomininGroup>> groupNeighborhood = new GridCellNgh<HomininGroup>(
                grid, currentLocation, HomininGroup.class,
                foragingRange, foragingRange).getNeighborhood(false); // TODO: was getTerritoryRange, ok?
        
        searchMerge:
		for (final GridCell<HomininGroup> act : groupNeighborhood) 
		{
			if (act.size() > 0) 
			{   
				// size is number of agents (all kinds) at this gridcell
				for (final Object object : act.items()) 
				{
	                if (object instanceof HomininGroup) 
	                {
	                	HomininGroup h = (HomininGroup) object;
	                	if (h.isMergingOk(getHomininType()))
	                	{
	                		// merging is ok
	                		hasMerged = true;
		                	h.sizeFertile += this.sizeFertile;
		                	h.sizePostFertile += this.sizePostFertile;
		                	h.sizePreFertile += this.sizePreFertile;
		                	h.isGroupHungry = true;
		                	
		    		        Constants.DEBUG_PRINT_LEVEL("Group merged - " + this, 15);
		                	// remove merged group
		    		        setAlive(false);

		                	// set the size to zero to avoid them being administrated as dead individuals
		                	sizeFertile = 0;
		                	sizePostFertile = 0;
		                	sizePreFertile = 0;
		                	
		                	HSUtils.incrementHomininGroupDeathCount(homininType, getSize(), Constants.CAUSE_OF_DEATH_MERGER);
		                	break searchMerge;
	                	}
	                }
		        }
			}
		}
		return hasMerged;
  	}
	
	/*
	 * Check if this group is still alive
	 */
    public boolean checkAlive() 
    {
    	/*
    	 * if our habitat turned into sea, remove the group
    	 */
		if (isAlive()) 
		{
			if (Constants.ZERO_MODEL_DISABLE_WATER == false)
			{
		    	if ((currentHabitatCell.getHabitatType() == Constants.HABITAT_TYPE_SEA) ||
		    			(currentHabitatCell.getClimateType() == Constants.CLIMATE_TYPE_SEA) )
		    	{
		    		Constants.DEBUG_PRINT_LEVEL("Die because of area flooding: " + this, 10);
		    		HSUtils.incrementHomininGroupDeathCount(homininType, getSize(), Constants.CAUSE_OF_DEATH_FLOODED);
			        setAlive(false);
		    	}
			}

			/*
	         * if temperature is too cold, remove the group
	         */
	    	if (currentHabitatCell.getTemperatureColdestMonth() <= minimumTemperature)
	    	{
	    		Constants.DEBUG_PRINT_LEVEL("Die because of extreme cold: " + this, 10);
	    		HSUtils.incrementHomininGroupDeathCount(homininType, getSize(), Constants.CAUSE_OF_DEATH_COLD);
		        setAlive(false);
	    	}
		}
		
    	/*
    	 * if almost all group members died, remove the group
    	 */
        if (isAlive() && (sizeFertile<=HomininSpaceContextBuilder.getGroupSizeFertileSmall())) 
        {
    		Constants.DEBUG_PRINT_LEVEL("Die because fertile group section is too small: " + this, 10);
    		HSUtils.incrementHomininGroupDeathCount(homininType, getSize(), Constants.CAUSE_OF_DEATH_SIZE);
	        setAlive(false);
        }
        
        
        if (!isAlive())
        {
        	if (dieCountdown-- < 0)
        	{
	        	// die - remove from simulation
	        	if (currentHabitatCell != null)
	        	{
	        		currentHabitatCell.addDeads((int) getSize());
	        		currentHabitatCell.setIsHomeRange(false);
	        	}
	    		HomininSpaceContextBuilder.deleteHomininGroup(this);
        	}
        }
        return isAlive();
	}    
    
    /*
     * Get system variables
     */
    public Grid<Object> getGrid() 
    {
		@SuppressWarnings("unchecked")
		final Grid<Object> grid = (Grid<Object>) ContextUtils.getContext(this)
				.getProjection(Constants.GRID_ID);

		if (null == grid) {
			throw new IllegalStateException("Cannot locate grid in context.");
		}

		return grid;
    }

    public int getTerritoryRange()
    {
		int territoryRange = 1;
		int climateType = currentHabitatCell.getClimateType();
		if (climateType == 1)  // polar
			territoryRange = 4;
		if (climateType == 2)  // boreal
			territoryRange = 3;
		if (climateType == 3)  // cool forest
			territoryRange = 3;
		if (climateType == 4)  // cool grass
			territoryRange = 3;
		if (climateType == 5)  // warm forest
			territoryRange = 2;
    	return territoryRange;
    }
    
    // angles are from 0 to 2*PI
	public static boolean isSameDirection( double angle1, double angle2 )
	{
		//angle1 += Math.PI; angle2 += Math.PI;
//		return ((angle1 < (angle2 + (0.25 * Math.PI))) && (angle1 > (angle2 - 0.25 * Math.PI)));
		return ((angle1 < (angle2 + (0.4 * Math.PI))) && (angle1 > (angle2 - 0.4 * Math.PI)));
	}
	
	/**
	 * Returns the angle of the vector from p0 to p1. The angle will be between -PI and PI, and then PI added.
	 * @return the angle (in radians) that p0p1 makes with the positive x-axis.
	 */
	public static double angle(int x1, int y1, int x2, int y2) 
	{
		double dx = x2 - x1;
		double dy = y2 - y1;
		
		double angle1;

		// method 1
		angle1 = Math.atan2(dy, dx) + Math.PI;  // atan2() returns range from -PI->PI, but moveByVector wants range 0->2PI;
/*		
		// method 2
		double angle2;
		if (dx > 0) { angle2 = (Math.PI*0.5) - Math.atan(dy/dx); }  else
			if (dx < 0) { angle2 = (Math.PI*1.5) - Math.atan(dy/dx); }   else
			if (dy > 0) { angle2 = Math.PI; }                          else
			if (dy < 0) { angle2 = Math.PI; }                        else
				angle2 = 0.0; // the 2 points are equal
		angle2 += Math.PI;
		
		Constants.DEBUG_PRINT_LEVEL( "Calculating angles: " + angle1 + " " + angle2, 5);
*/		
		// also possible, maybe better ???
		//double angle3 = SpatialMath.calcAngleFor2DMovement(HomininSpaceContextBuilder.theSpace, myPoint, otherPoint);
		
		return angle1; //
	}
    
	/*
	 * Check to see if merging is ok for this group
	 *    ignore hominingroup type for now
	 */
	public boolean isMergingOk( int aType )
	{
		boolean res = true;
		
		if (isHungry() || !isAlive())
		{
			// already hungry, this group will not accept more individuals
			res = false;
		}
		else
		if (getSize() > HomininSpaceContextBuilder.getGroupSizeLarge())
		{
			// maximum size reached
			res = false;
		}
		
		return res;
	}
	
    /*
     * Getters and setters
     */
    public double getSizePreFertile() {
        return sizePreFertile;
    }
    public double getSizeFertile() {
        return sizeFertile;
    }
    public double getSizePostFertile() {
        return sizePostFertile;
    }
    public double getSizeType0() {
    	double res = 0;
    	if (this.homininType == Constants.CULTURAL_TYPE_0)
    		res = getSize();
    	return res;
    }
    public double getSizeType1() {
    	double res = 0;
    	if (this.homininType == Constants.CULTURAL_TYPE_1)
    		res = getSize();
    	return res;
    }
    public double getSizeType2() {
    	double res = 0;
    	if (this.homininType == Constants.CULTURAL_TYPE_2)
    		res = getSize();
    	return res;
    }

	/*
	 * The parameter construction allows the user to change this variable while running the application
	 *   by double clicking the agent
	 */
    //@Parameter(displayName = "Group Size", usageName = "size")
    public double getSize() {
        return sizePreFertile + sizeFertile + sizePostFertile;
    }
    public int getRoundedSize() {
        return (int) Math.round(sizePreFertile + sizeFertile + sizePostFertile + 0.5);
    }
	public void setSize(final double aSizePreFertile, final double aSizeFertile, final double aSizePostFertile) {
		this.sizePreFertile = aSizePreFertile;
		this.sizeFertile = aSizeFertile;
		this.sizePostFertile = aSizePostFertile;
	}
	public double getCaloriesRequired()
	{
		double caloriesRequired = 
				consumptionPreFertile * this.sizePreFertile * 365 +
				consumptionFertile * this.sizeFertile * 365 +
				consumptionPostFertile * this.sizePostFertile * 365;
		return caloriesRequired;
	}
    public double getAge() {
        return age;
    }
    public void setAge( double anAge)
    {
    	this.age = (int) anAge;
    }

    public int getPosX() {
		posX = currentLocation.getX();
        return posX;
    }
    public int getPosY() {
		posY = currentLocation.getY();
        return posY;
    }
	public void setPosXY(final int x, final int y, final int oldPosX, final int oldPosY) {
		grid.moveTo(this, x, y);
		this.posX = x;
		this.posY = y;
		currentLocation = grid.getLocation(this);
		
		// find current habitatcell
        final Iterable<Object> objects = grid.getObjectsAt(x,y);
        for (final Object object : objects) 
        {
        	if (object instanceof HabitatCell) 
        	{
        		HabitatCell h = (HabitatCell) object;
        		currentHabitatCell = h;
        		break;
        	}
        }
	}	

    public int getHomininType() {
        return homininType;
    }
	public void setHomininType(final int aType) {
		homininType = aType;
	}
    public int getForagingRange() {
        return foragingRange;
    }

    public int getId() {
        return id;
    }
    public boolean isAlive()
    {
    	return isGroupAlive;
    }
    public void setAlive( final boolean aStatus)
    {
    	isGroupAlive = aStatus;
    	if (aStatus == false)
    	{
    		dieCountdown = 5;   // 5 turns till group is removed
    	}
    }
    public boolean isHungry()
    {
    	return isGroupHungry;
    }
    public boolean isSmall()
    {
    	return isGroupSmall;
    }
   
	public static double getGlobalTemperature()
	{
		return HomininSpaceContextBuilder.getGlobalTemperatureOffset();
	}
	public static double getGlobalSeaLevel()
	{
		return HomininSpaceContextBuilder.getGlobalSeaLevel();
	}
	public static double getHungryGroupPercentage()
	{
		return HomininSpaceContextBuilder.getHungryGroupPercentage();
	}
	public static double getHungryHomininPercentage()
	{
		return HomininSpaceContextBuilder.getHungryHomininPercentage();
	}
	public void setGroupSettled( boolean value )
	{
		isGroupSettled = value;
	}
    public boolean isSettled() {
        return isGroupSettled;
    }
	public void setCanSettle( boolean value )
	{
		canSettle = value;
	}
    public double getIconSize()
    {
    	return 18;
    }
    
    // color codes for the different hominingroup types
    public double getIconColorRed()
    {
    	if (this.getHomininType() == 0)
    		return 115;
    	else
        	if (this.getHomininType() == 1)
        		return 255;
        	else
            	if (this.getHomininType() == 2)
            		return 255;
            	else
            		return 0;
    }
    public double getIconColorGreen()
    {
    	if (this.getHomininType() == 0)
    		return 242;
    	else
        	if (this.getHomininType() == 1)
        		return 0;
        	else
            	if (this.getHomininType() == 2)
            		return 255;
            	else
        	return 0;
    }
    public double getIconColorBlue()
    {
    	if (this.getHomininType() == 0)
    		return 0;
    	else
        	if (this.getHomininType() == 1)
        		return 0;
        	else
            	if (this.getHomininType() == 2)
            		return 255;
            	else
        	return 0;
    }
    
    private String printSize()
    {
        return String.format("size=(%.2f; %.2f; %.2f)", this.sizePreFertile, this.sizeFertile, this.sizePostFertile);
    }
    
	@Override
	public String toString() {
		// Override default Java implementation just to have a more meaningful representation
        return String.format("HomininGroup %d at (%d %d), %s cal:%10.0f (age=%d) (range=%d) (hungry=%d deaths:%.2f %.2f %.2f)", id, posX, posY, printSize(), getCaloriesRequired(), this.age, this.foragingRange, yearsHungry, deathsPreFertile, deathsFertile, deathsPostFertile);
	}
}