/*
 Copyright 2011-2018 Fulco Scherjon

This file is part of the HomininSpace modelling and simulation system.

HomininSpace is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, any version 3.
HomininSpace is distributed WITHOUT ANY WARRANTY under the GNU General Public License.
HomininSpace comes with a copy of this GNU General Public License.  If not, see <http://www.gnu.org/licenses/>.

*/

package HomininSpace.common;

import java.io.Serializable;
import java.util.Comparator;

/*
 * this class stores the data as it is retrieved from an input file. Each Habitat cell has an attached CellData object 
 *   if that was present in the input file. This CellData object keeps the original data.
 *   
 */
public class CellData {
    private final int x, y;
    private double height;
    private double depth;
    private double currentPrecipitation;		// average precipitation through the year today
    private double currentTemperature;         // average temperature through the year today
    private double currentColdestMonthTemperature;		// min temperature in the coldest month
    private double currentWarmestMonthTemperature;         // max temperature in the warmest month
    private double LGMPrecipitation;		// average precipitation through the year during the LGM
	private double LGMTemperature;         // average temperature through the year during the LGM
    private double LGMColdestMonthTemperature;		// min temperature in the coldest month
    private double LGMWarmestMonthTemperature;         // max temperature in the warmest month
    
    private int typeCell = Constants.HABITAT_TYPE_PLAIN;

    public static final CellDataXComparator CELL_DATA_X_COMPARATOR = new CellDataXComparator();
    public static final CellDataYComparator CELL_DATA_Y_COMPARATOR = new CellDataYComparator();

    private static final class CellDataXComparator implements
                    Comparator<CellData>, Serializable {
            private static final long serialVersionUID = 4149351630089726905L;

            @Override
            public int compare(final CellData o1, final CellData o2) {
                    return o1.getX() - o2.getX();
            }
    }

    private static final class CellDataYComparator implements
                    Comparator<CellData>, Serializable {
            private static final long serialVersionUID = -5955739679291874417L;

            @Override
            public int compare(final CellData o1, final CellData o2) {
                    return o1.getY() - o2.getY();
            }
    }

    public CellData(final int x, final int y) {
            this.x = x;
            this.y = y;
    }

    public int getX() {
            return x;
    }

    public int getY() {
        return y;
}

    public double linearInterpolationWithOffset( final double v1, final double v2)
    {
    	double globalTempOffset = HSUtils.getTemperatureOffset();
    	return v1 + ((v2-v1) * (-globalTempOffset / 15.936)); // minimum LGM temp
    }
    
    public double getTemperatureColdestMonthInterpolated()
    {
    	double t1 = currentColdestMonthTemperature;
    	double t2 = LGMColdestMonthTemperature;
    	
    	return linearInterpolationWithOffset(t1, t2); 
    }
    
    public double getTemperatureWarmestMonthInterpolated()
    {
    	double t1 = currentWarmestMonthTemperature;
    	double t2 = LGMWarmestMonthTemperature;
    	
    	return linearInterpolationWithOffset(t1, t2); 
    }
    
    public double getTemperatureInterpolated() {
    	final double t1 = currentTemperature;
    	final double t2 = LGMTemperature;
    	final double res = linearInterpolationWithOffset(t1, t2); 
    	
    	return res; 
    }
    
    public double getPrecipitationInterpolated() {
    	double p1 = currentPrecipitation;
    	double p2 = LGMPrecipitation;
    	double res = linearInterpolationWithOffset(p1, p2);
    	
    	if (res < 0)
    	{
    		Constants.DEBUG_PRINT_LEVEL("Precip < 0: p1=" + p1 + " p2=" + p2 , 5);
    	}
    	
    	return res; 
    }
    public void setHeight( double aHeight )
    {
    	this.height = aHeight;
    }
    public double getHeight() {
            return height;
    }
    public void setType( int aType )
    {
    	this.typeCell = aType;
    }
    public int getType() {
            return typeCell;
    }

    public double getDepth() {
		return depth;
	}

	public void setDepth(double depth) {
		this.depth = depth;
	}

	public double getCurrentPrecipitation() {
		return currentPrecipitation;
	}

	public void setCurrentPrecipitation(double currentPrecipitation) {
		this.currentPrecipitation = currentPrecipitation;
	}

	public double getCurrentTemperature() {
		return currentTemperature;
	}

	public void setCurrentTemperature(double currentTemperature) {
		this.currentTemperature = currentTemperature;
	}

	public double getCurrentColdestMonthTemperature() {
		return currentColdestMonthTemperature;
	}

	public void setCurrentColdestMonthTemperature(
			double currentColdestMonthTemperature) {
		this.currentColdestMonthTemperature = currentColdestMonthTemperature;
	}

	public double getCurrentWarmestMonthTemperature() {
		return currentWarmestMonthTemperature;
	}

	public void setCurrentWarmestMonthTemperature(
			double currentWarmestMonthTemperature) {
		this.currentWarmestMonthTemperature = currentWarmestMonthTemperature;
	}

	public double getLGMColdestMonthTemperature() {
		return LGMColdestMonthTemperature;
	}

	public void setLGMColdestMonthTemperature(double lGMColdestMonthTemperature) {
		LGMColdestMonthTemperature = lGMColdestMonthTemperature;
	}

	public double getLGMWarmestMonthTemperature() {
		return LGMWarmestMonthTemperature;
	}

	public void setLGMWarmestMonthTemperature(double lGMWarmestMonthTemperature) {
		LGMWarmestMonthTemperature = lGMWarmestMonthTemperature;
	}

	public double getLGMPrecipitation() {
		return LGMPrecipitation;
	}

	public void setLGMPrecipitation(double lGMPrecipitation) {
		LGMPrecipitation = lGMPrecipitation;
	}

	public double getLGMTemperature() {
		return LGMTemperature;
	}

	public void setLGMTemperature(double lGMTemperature) {
		LGMTemperature = lGMTemperature;
	}

    @Override
    public String toString() {
            return String.format("CellData [x=%d, y=%d]", x, y);
    }

}