/*
 Copyright 2011-2018 Fulco Scherjon

This file is part of the HomininSpace modelling and simulation system.

HomininSpace is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, any version 3.
HomininSpace is distributed WITHOUT ANY WARRANTY under the GNU General Public License.
HomininSpace comes with a copy of this GNU General Public License.  If not, see <http://www.gnu.org/licenses/>.

*/


package HomininSpace.common;

import HomininSpace.HomininSpaceContextBuilder;

public class Constants {
	
	/*
	 * Model System settings
	 */
	public static final String 					VERSION 					= "1.5.223";
	
	// if ENERGY_CONTINUOUS = true the model uses continuous energy levels (McNaughton), otherwise Kelly/Binford habitat reconstructions
	//    McNaughton gives a more even spread of energy, Kelly/Binford give patches of energy
	public static boolean						ENERGY_CONTINUOUS = true;

	// if STATIC_DISPERSAL = true groups will settle where possible
	public static boolean						STATIC_DISPERSAL = true;
	
	// constant for implementing increasing levels of zero models as baseline models (reference models)
	public static boolean						ZERO_MODEL_RANDOM_WALK = false;  // when true groups move randomly through the landscape
	public static boolean						ZERO_MODEL_DISABLE_WATER = false;  // when false grid cells under water are inaccessible, when true there is no water in the model
	public static boolean						ZERO_MODEL_DISABLE_ENERGY = false;  // when false grid cells have energy from climate parameters, when true there is a uniform initial energy distribution
	
	// if USE_MAXIMUM_FORAGING_RANGE = true the parameter value for maximum foraging range is enforced; if not the range is unlimited (adaptive)
	public static boolean						USE_MAXIMUM_FORAGING_RANGE = true;
	
	/*
	 *  Debugging information
	 * 		Debuglevel indicate how much information is given, the higher the number the more info
	 */
	public static final boolean 				DEBUG_MODE 	= true;   	// if true debug information is send to the console
	public static int 							DEBUG_LEVEL = 50;   	// if DEBUG_MODE=true all information levels less than this are printed

    /*
     * Simulation schedule parameters
     */
	public static final double 					START_TIME_STEP 		= 131;   // in kya BP
	public static final double 					END_TIME_STEP 			= 50;
	
	/*
	 * Parameter names (as defined in user interface) (deprecated)
	 */
    public static final String 					PARAMETER_ID_DEFAULT_LEVEL = "defaultLevel";
    public static final String 					PARAMETER_ID_SETTINGS_INDEX = "settingsIndex";
    public static final String 					PARAMETER_ID_TICK_COUNT 	= "tickCount";
    public static final String 					PARAMETER_ID_STARTING_TIME_STEP = "startingTimeStep";
    public static final String 					PARAMETER_ID_TIME_STEP_INCREMENT = "timeStepIncrement";
    public static final String 					PARAMETER_ID_HOMININ_COUNT = "homininCount";
    public static final String 					PARAMETER_ID_FACTORY_HUNGER_PERCENTAGE = "factoryHungryPercentage";
    public static final String 					PARAMETER_ID_PAUZE_BEFORE_END = "pauzeBeforeEnd";
    public static final String 					PARAMETER_ID_INITIAL_IBERIA_GROUPS = "initialIberiaGroupCount";
    public static final String 					PARAMETER_ID_INITIAL_ITALY_GROUPS = "initialItalyGroupCount";
    public static final String 					PARAMETER_ID_MINIMUM_TEMPERATURE = "minimumTemperature";
    public static final String 					PARAMETER_ID_BIRTHS_PERCENTAGE = "birthsPercentage";
    public static final String 					PARAMETER_ID_OPEN_BORDERS = "openBorders";
    public static final String 					PARAMETER_ID_STATIC_DISPERSAL = "staticDispersal";
    public static final String 					PARAMETER_ID_MODEL_DATA_DIRECTORY = "modelDataDirectory";

    /*
     * MIS Isotope stages
     */
	//
	// http://en.wikipedia.org/wiki/Marine_isotope_stage	
	// All numbers are taken from Aitken & Stokes, Table 1.4, except for the sub-stages of MIS 5, which are from Wright's Table 1.1.
	//    Aitken, Martin J and Stokes, Stephen, in Taylor, Royal Ervin Taylor and Aitken, Martin Jim (eds), Chronometric dating in archaeology, Chapter 1, 1997, Birkhuser
	//    Wright, James D., "Global Climate Change in Marine Stable Isotope Records", in Noller, Jay S., Sowers, Janet M., Lettis, William R. (eds), Quaternary geochronology: methods and applications, 2000, American Geophysical Union
	// 	
	
	public static final double 					START_MIS6		 = 190000;   // in kya BP
	public static final double 					START_MIS5e		 = 130115;   // in kya BP
	public static final double 					START_MIS5d		 = 115105;   // in kya BP
	public static final double 					START_MIS5c		 = 105920;   // in kya BP
	public static final double 					START_MIS5b		 =  92840;   // in kya BP
	public static final double 					START_MIS5a		 =  84740;   // in kya BP
	public static final double 					START_MIS4		 =  71000;   // in kya BP
	public static final double 					START_MIS3		 =  60000;   // in kya BP

	/*
	 * Checkpoint types
	 *   General: stores presence, and matches against archaeololgy
	 *   Monitoring: stores presence, but has no archaeology associated
	 *   Startingpoint: creates one group at startup, monitors presence
	 *   Climate: stores climate parameters for each time step
	 */
	public enum CheckpointType {
		CHECKPOINT_TYPE_GENERAL_CST(1), 
		CHECKPOINT_TYPE_MONITORING_CST(2), 
		CHECKPOINT_TYPE_STARTINGPOINT_CST(3), 
		CHECKPOINT_TYPE_CLIMATE_CST(4);
		
		private final int ctype;
		
		private CheckpointType(int atype)
		{
			this.ctype = atype;
		}
	}

	/*
	 * Habitat constants
	 */
    public static final double 					SEA_LEVEL_HEIGHT 			= 0.0;  // land below this heigth is considered sea
    public static final double 					PLAIN_HEIGHT 				= 300.0;  // land below this heigth is considered plain
    public static final double 					HILL_HEIGHT 				= 600.0;  // land below this heigth is considered hill  (above is mountain)
    
    /* types of habitat: water, beach, hill, mountain, plain   */
    public static final int 					HABITAT_TYPE_SEA 			= 0;
    public static final int 					HABITAT_TYPE_LAKE 			= 1;
    public static final int 					HABITAT_TYPE_BEACH 			= 2;
    public static final int 					HABITAT_TYPE_PLAIN			= 3;
    public static final int 					HABITAT_TYPE_HILL 			= 4;
    public static final int 					HABITAT_TYPE_MOUNTAIN 		= 5;
    
    // types of climate
    public static final int 					CLIMATE_TYPE_SEA 			= 0;
    public static final int 					CLIMATE_TYPE_POLAR 			= 1;
    public static final int 					CLIMATE_TYPE_BOREAL			= 2;
    public static final int 					CLIMATE_TYPE_COOL_FOREST	= 3;
    public static final int 					CLIMATE_TYPE_COOL_GRASS		= 4;
    public static final int 					CLIMATE_TYPE_WARM_FOREST	= 5;
    public static final int 					CLIMATE_TYPE_WARM_GRASS		= 6;
    public static final int 					CLIMATE_TYPE_SUBTROPICAL	= 7;
    public static final int 					CLIMATE_TYPE_TROPICAL		= 8;
    public static final int 					CLIMATE_TYPE_EQUATORIAL		= 9;

    /*
     * Hominin group defining constants
     */
    public static final double 					INITIAL_SIZE_PREFERTILE 		= 8.0;
    public static final double 					INITIAL_SIZE_FERTILE 			= 11.0;
    public static final double 					INITIAL_SIZE_POSTFERTILE 		= 6.0;
    
    /*
     * Types of hominin group sources
     *   0 = initial groups, present at start of the simulation
     *   1 = coming from Iberia
     *   2 = coming from Italy
     */
    public static final int 					TYPES_OF_HOMININS 				= 3;
    public static final int 					CULTURAL_TYPE_0					= 0;
    public static final int 					CULTURAL_TYPE_1					= 1;
    public static final int 					CULTURAL_TYPE_2					= 2;
    
    /*
     * Causes of death
     */
    public static final int 					CAUSE_OF_DEATH_BORDER			= 1;
    public static final int 					CAUSE_OF_DEATH_BORDER_IBERIA	= 2;
    public static final int 					CAUSE_OF_DEATH_COLD				= 3;
    public static final int 					CAUSE_OF_DEATH_SIZE				= 4;
    public static final int 					CAUSE_OF_DEATH_FLOODED			= 5;
    public static final int 					CAUSE_OF_DEATH_MERGER			= 6;
    public static final int 					CAUSE_OF_DEATH_HUNGER			= 7;
    
    /*
     * Factory constants
     */
    public static final int 					FACTORY_SCANNING_RANGE			= 30;
    
    /*
     * System constants
     */
    public static final String 					CONTEXT_ID 					= "HomininSpace";
    public static final String 					SPACE_ID 					= "space";
    public static final String 					GRID_ID						= "grid";
    public static final String 					VISIBLE_VALUE_LAYER_ID 		= "visibleValueLayer";
    
//    public static final String					MODEL_DATA_DIRECTORY 		= "D:\\Repast Workspace\\HomininSpace\\";
    public static final String					MODEL_DATA_DIRECTORY 		= ".\\";
    
	/*
	 * Visual representation of the simulation: describes how and what the map shows during the simulation 
	 * TODO: enumeration
	 */
	public static int 							VISIBLE_LAYER_TOPOGRAPHY = 1;
	public static int 							VISIBLE_LAYER_CLIMATE = 2;
	public static int 							VISIBLE_LAYER_ENERGY = 3;
	public static int 							VISIBLE_LAYER_VISITED = 4;
	public static int 							VISIBLE_LAYER_DEAD = 5;
 
	// number of simulation steps the topography is displayed to the user
	public static int 							VISIBLE_LAYER_TOPOGRAPHY_TURN_COUNT = -1;
	
	/*
	 * Visual layer that is shown during the simulation
	 */
												// WHILE_RUNNING can be: VISIBLE_LAYER_CLIMATE, VISIBLE_LAYER_ENERGY, VISIBLE_LAYER_TOPOGRAPHY
	public static int							VISIBLE_LAYER_WHILE_RUNNING = VISIBLE_LAYER_ENERGY;
	
	/*
	 * Visible layer that is shown when the simulation is finished
	 */
												// AT_END can be: VISIBLE_LAYER_VISITED, VISIBLE_LAYER_DEAD
	public static int							VISIBLE_LAYER_AT_END = VISIBLE_LAYER_VISITED;
	
    
	// Functions
	public static void DEBUG_PRINT( String s )
	{
		if (Constants.DEBUG_MODE) {
			System.out.println("[" + HomininSpaceContextBuilder.getSimulationId() + "] " + HomininSpaceContextBuilder.getYear() + " (" + Math.round(HomininSpaceContextBuilder.getTickCount()) + "):" + s);
		}
	}
	
	public static void DEBUG_PRINT_LEVEL( String s, int aLevel )
	{
		if (Constants.DEBUG_MODE) {
			if (aLevel < DEBUG_LEVEL) {
				DEBUG_PRINT(s);
			}
		}
	}


    /**
	 * Hidden constructor to ensure no instances are created.
	 */
	private Constants() {
		;
	}
}

